import os
import sys
import copy
import numpy as np
import torch
import cv2

from core.config import cfg
from core.sim_module import SimilarityRCNN
from models.siam_rcnn_rpn import SiamRPN
from models.siam_rcnn_head import SiamRCNNHead
from rpn.generate_anchors import generate_anchors_on_response_map
from utils.model_weights import load_weights
import utils.boxes as ubox
import utils.crop as ucrop
import utils.rpn_utils as urpn
from utils.nms import nms_numpy


class SiamRCNN_Tracker(object):

    def __init__(self, model_device):

        self.model_device = model_device

        self.net = SiamRPN(cfg, is_training=False)
        load_weights(self.net, cfg.SIAMRPN.WEIGHTS_FILE)
        self.rcnn = SiamRCNNHead(cfg, is_training=False)
        load_weights(self.rcnn, cfg.RCNN.WEIGHTS_FILE)

        self.net.to(model_device)
        self.rcnn.to(model_device)
        self.net.eval()
        self.rcnn.eval()
        
        # construct similarity module
        self.sim_module = SimilarityRCNN(self.rcnn, self.model_device)

        self.ar_range = np.array(cfg.SIAMRPN.ASPECT_RATIO_RANGE)
        self.z_size = cfg.INFERENCE.Z_IMAGE_SIZE
        self.x_size = cfg.INFERENCE.X_IMAGE_SIZE
        self.z_res_size = cfg.INFERENCE.Z_RESPONSE_SIZE
        self.x_res_size = cfg.INFERENCE.X_RESPONSE_SIZE

        # generate anchors
        self.anchors = generate_anchors_on_response_map(self.z_size, 
                                                        self.x_size, 
                                                        self.z_res_size, 
                                                        self.x_res_size, 
                                                        cfg.SIAMRPN.STRIDE,
                                                        cfg.SIAMRPN.ANCHOR_RATIOS,
                                                        cfg.SIAMRPN.ANCHOR_SCALES)
    
    def init_tracker(self, first_frame_img, current_box):
        ''' Init the tracker '''
        with torch.no_grad():
            
            search_img_tensor, z_crop_box, x_crop_box, scale_x = self.generate_search_images(first_frame_img, current_box)
            
            # calculate box aspect ratio
            xc, yc, current_w, current_h = ubox.xyxy_to_xcycwh(current_box)
            z_box_ar = np.abs(current_w) / (np.abs(current_h) + 0.0001)
            z_box_ar_index = int(np.searchsorted(self.ar_range, z_box_ar))  

            z_out = self.net(z_image=search_img_tensor, x_image=None, ar_index = z_box_ar_index)

            z_box_in_search_img = np.array([0.] + current_box, dtype=np.float32)
            z_box_in_search_img[1:5:2] -= x_crop_box[0]
            z_box_in_search_img[2:5:2] -= x_crop_box[1]
            z_box_in_search_img[1:5] *= scale_x
            # get z_features of the first frame
            z_rois = torch.tensor(z_box_in_search_img).type_as(search_img_tensor).view(1, 5)
            z_feats = self.rcnn.feat_extractor(z_out, z_rois)
            self.sim_module.init_template(z_feats)

    def predict_next_frame(self, next_frame_img, current_box, vis=False):
        xc, yc, current_w, current_h = ubox.xyxy_to_xcycwh(current_box)
        current_box_xyxy = np.array(current_box, dtype=np.float32)
        
        im_height = next_frame_img.size(2)
        im_width = next_frame_img.size(3)

        # SiamRPN
        with torch.no_grad():
            search_imgs_tensor, z_crop_box, x_crop_box, scale_x = \
                self.generate_search_images(next_frame_img, current_box)
            det_boxes, z_out, x_out = self.generate_rpn_boxes(search_imgs_tensor, x_crop_box, scale_x)
            # det_boxes: [N x 9], in order of [img_x1, img_y1, img_x2 img_y2, rpn_score, x1, y1, x2, y2]
        
        proposed_boxes = self.generate_proposal_boxes(det_boxes, current_box_xyxy)
        
        # sim_scores, sim_boxes = self.get_similarity_from_rcnn(x_out, proposed_boxes[:, 5:9].copy())
        sim_scores, sim_boxes = self.sim_module.get_similarity(x_out, proposed_boxes[:, 5:9].copy())

        if cfg.INFERENCE.PRE_COSINE_WIN:
            # post-processing for proposals
            win_scores = self.calculate_window_scores(proposed_boxes[:, 5:9], x_crop_box, scale_x, on_image=False)

        if cfg.INFERENCE.ENABLE_VERIFICATION_BBOX_REFINE:
            # transform to original image domain
            sim_boxes_img = sim_boxes / scale_x
            sim_boxes_img[:, 0:4:2] += x_crop_box[0]
            sim_boxes_img[:, 1:4:2] += x_crop_box[1]

            # calculate the ensemble weights
            if cfg.INFERENCE.ENABLE_INDEPENDENT_BBOX_WEIGHT:
                _w = 1.0 - cfg.INFERENCE.VERIFICATION_BBOX_INFLUENCE
            else:
                v_influence = cfg.INFERENCE.VERIFICATION_BBOX_INFLUENCE
                sum_scores = np.clip(proposed_boxes[:, 4] + sim_scores * v_influence, a_min=0.00001, a_max=99999.)
                _w = (proposed_boxes[:, 4] / sum_scores)[:, np.newaxis]
            proposed_boxes[:, :4] = proposed_boxes[:, :4] * _w + sim_boxes_img[:, :4] * (1. - _w)
            proposed_boxes[:, 5:9] = proposed_boxes[:, 5:9] * _w + sim_boxes[:, :4] * (1. - _w)
        
        if not cfg.INFERENCE.PRE_COSINE_WIN:
            # post-processing for proposals
            win_scores = self.calculate_window_scores(proposed_boxes[:, 5:9], x_crop_box, scale_x, on_image=False)
        
        rpn_scores = proposed_boxes[:, 4]
        pred_scores = rpn_scores * (1 - cfg.INFERENCE.VERIFICATION_INFLUENCE) + \
                        sim_scores * cfg.INFERENCE.VERIFICATION_INFLUENCE

        final_scores = pred_scores * (1 - cfg.INFERENCE.WINDOWS_INFLUENCE) + \
                            win_scores * cfg.INFERENCE.WINDOWS_INFLUENCE
                
        max_score_idx = np.argmax(final_scores)
        best_box = proposed_boxes[max_score_idx, :4].tolist()
        best_score = pred_scores[max_score_idx]

        if best_score < cfg.INFERENCE.MAX_SCORE_THRESHOLD:
            best_box = current_box
            lr = 1.0 * cfg.INFERENCE.LR
        else:
            lr = best_score * cfg.INFERENCE.LR
        
        best_box  = ubox.xyxy_to_xcycwh(best_box)
        best_box[2] = best_box[2] * lr + current_w * (1 - lr)
        best_box[3] = best_box[3] * lr + current_h * (1 - lr)

        best_box[2] = max(10, best_box[2])
        best_box[3] = max(10, best_box[3])
        best_box[0] = max(0, min(best_box[0], im_width - (best_box[2]) / 2.))
        best_box[1] = max(0, min(best_box[1], im_height - (best_box[3]) / 2.))

        best_box = ubox.xcycwh_to_xyxy(best_box)
        
        if cfg.INFERENCE.ENABLE_UPDATE:
            self.sim_module.update(proposed_boxes[:, :4],
                                proposed_boxes[:, 5:9], 
                                proposed_boxes[:, 4], 
                                win_scores[:], 
                                max_score_idx)

        return best_box
        

    def generate_search_images(self, next_frame_img, current_box):
        ''' Given an input image and target box, generate the images of search regions.

        Args:
            next_frame_img: Input image, a numpy array in shape of [H, W, 3]
            current_box: Bounding box location, torch.Tensor in shape of [1, 4]

        Returns:
            search_imgs: images of search regions, a Torch tensor in shape of [N, 3, X_IMAGE_SIZE, IMAGE_SIZE]
            scale_x_list: scale factors of search regions, a list of float numbers.
            box_list: search region box list, in shape of [x1, y1, x2, y2]
        '''
        z_box, x_box, scale_x = ucrop.generate_search_region_boxes(current_box, 
                                            size_z=self.z_size, 
                                            size_x=self.x_size,
                                            context_amount=cfg.SIAMRPN.Z_CONTEXT_AMOUNT)
        search_imgs_tensor = \
                ucrop.crop_with_boxes(next_frame_img, x_box, self.x_size, self.x_size)

        return search_imgs_tensor, z_box, x_box, scale_x


    def calculate_window_scores(self, det_boxes, x_crop_box, scale_x, on_image=True):
        if on_image:
            det_boxes_on_x = det_boxes.copy()
            det_boxes_on_x[:, 0:4:2] -= x_crop_box[0]
            det_boxes_on_x[:, 1:4:2] -= x_crop_box[1]
            det_boxes_on_x[:, 0:4] *= scale_x
        else:
            det_boxes_on_x = det_boxes
        det_boxes_on_x_xcycwh = ubox.xyxy_to_xcycwh(det_boxes_on_x[:, :4])
        center_dist = det_boxes_on_x_xcycwh[:, 0:2] / (self.x_size - 1)
        center_dist = np.maximum(np.minimum(center_dist, 1.0), 0.0)
        win_scores_2d = 0.5 - 0.5 * np.cos(2. * np.pi * center_dist)
        win_scores = win_scores_2d[:, 0] * win_scores_2d[:, 1]
        return win_scores


    def generate_rpn_boxes(self, search_imgs_tensor, x_crop_box, scale_x):
        z_out, x_out, rpn_cls_logit, rpn_bbox_delta = self.net(z_image=None, x_image=search_imgs_tensor, ar_index=None)            
        rpn_cls_prob = \
            urpn.rpn_logit_to_prob(rpn_cls_logit, cfg.SIAMRPN.RPN_HEAD_LOSS == 'softmax')
        rpn_cls_prob = rpn_cls_prob[:, rpn_cls_prob.size(1) // 2:, :, :]
        
        rpn_bbox_delta = rpn_bbox_delta.data.cpu().numpy()[0].transpose((1, 2, 0)).reshape(-1, 4)
        rpn_score = rpn_cls_prob.data.cpu().numpy()[0].transpose((1, 2, 0)).flatten()
        rpn_bbox_pred = ubox.bbox_transform(self.anchors, rpn_bbox_delta)
        
        # transform to original image domain
        rpn_bbox_pred_img = rpn_bbox_pred / scale_x
        rpn_bbox_pred_img[:, 0:4:2] += x_crop_box[0]
        rpn_bbox_pred_img[:, 1:4:2] += x_crop_box[1]

        det_boxes = np.hstack((rpn_bbox_pred_img, rpn_score[:, np.newaxis], rpn_bbox_pred))
        return det_boxes, z_out, x_out


    def generate_proposal_boxes(self, rpn_boxes, current_box_xyxy):
        # first step, extract some candidates which are closest to current_box
        num_near_candidates = cfg.INFERENCE.NUM_NEAR_CANDIDATES
        num_candidates = cfg.INFERENCE.NUM_CANDIDATES

        if num_near_candidates > 0:
            overlaps = ubox.bbox_overlaps(np.ascontiguousarray(rpn_boxes[:, :4], dtype=np.float32), 
                    current_box_xyxy.reshape(1, 4)).flatten()
            near_inds = np.where(overlaps > 0.80)[0]
            if len(near_inds) > 0 and len(near_inds) > num_near_candidates:
                ov_sorted_inds = near_inds[np.argsort(rpn_boxes[near_inds, 4])[::-1]]
                near_boxes = rpn_boxes[ov_sorted_inds[:num_near_candidates], :]
                
                other_inds = np.where(overlaps <= 0.80)[0]
                other_inds = np.concatenate((ov_sorted_inds[num_near_candidates:], other_inds))
                proposed_boxes = rpn_boxes[other_inds, :]
            else:
                ov_sorted_inds =np.argsort(overlaps)[::-1]
                near_boxes = rpn_boxes[ov_sorted_inds[:num_near_candidates], :]
                proposed_boxes = rpn_boxes[ov_sorted_inds[num_near_candidates:], :]
        else:
            proposed_boxes = rpn_boxes
        
        # filter some box whose score is really low
        max_score = np.max(proposed_boxes[:, 4])
        kept_inds = np.where(proposed_boxes[:, 4] >= cfg.INFERENCE.SCORE_THRESH_FACTOR * max_score)[0]
        proposed_boxes = proposed_boxes[kept_inds, :]

        if proposed_boxes.shape[0] > 1:
            kept_inds = nms_numpy(np.ascontiguousarray(proposed_boxes[:, :5], dtype=np.float32), cfg.INFERENCE.NMS_THRESHOLD)
            proposed_boxes = proposed_boxes[kept_inds, :]
            sort_inds = np.argsort(proposed_boxes[:, 4])[::-1]
            proposed_boxes = proposed_boxes[sort_inds, :]
        
            _k = min(proposed_boxes.shape[0], num_candidates - num_near_candidates)
            proposed_boxes = proposed_boxes[:_k, :]

        if num_near_candidates > 0:
            proposed_boxes = np.vstack((proposed_boxes, near_boxes))
        
        return proposed_boxes



