import torch
import numpy as np
import torch.nn.functional as F
from core.config import cfg
import utils.boxes as ubox

MAX_FRAMES = 4000

class SimilarityRCNN(object):

    def __init__(self, model, device):
        
        self.model = model
        self.model_device = device

        self.z_size = cfg.RCNN.POOL_SIZE[0]
        self.z_channels = cfg.RCNN.POOL_CHANNEL

        self.max_num_temps = cfg.INFERENCE.MAX_NUM_TEMPS
        # setup templates
        self.temp_embeddings = torch.zeros(self.max_num_temps, 
                                               self.z_channels,
                                               self.z_size,
                                               self.z_size,
                                               device=self.model_device,
                                               dtype=torch.float32)
        self.temp_weights = np.zeros((self.max_num_temps, ), dtype=np.float32)
        self.temp_inds = []
        self.temp_enable_update = np.zeros((self.max_num_temps, ), dtype=np.bool)
        self.num_temps = 0

        # setup negative templates
        self.use_negative = cfg.INFERENCE.ENABLE_NEGATIVE_UPDATE
        if self.use_negative:
            
            self.max_num_neg_temps = cfg.INFERENCE.MAX_NUM_NEG_TEMPS
            self.neg_temp_embeddings = torch.zeros(self.max_num_neg_temps, 
                                               self.z_channels,
                                               self.z_size,
                                               self.z_size,
                                               device=self.model_device,
                                               dtype=torch.float32)
            self.neg_temp_weights = np.zeros((self.max_num_neg_temps, ), dtype=np.float32)
            self.neg_temp_inds = []
            self.neg_temp_enable_update = np.zeros((self.max_num_neg_temps, ), dtype=np.bool)
            self.num_neg_temps = 0
            

        # setup update rule
        update_rule_dict = {
            'naive': 0,
            'pos_only': 1,
        }
        self.update_rule = update_rule_dict[cfg.INFERENCE.UPDATE_TYPE]
        # cahced box information, in order of 
        # [xc, yc, w, h, rpn_score, win_score, max_sim_score, avg_sim_score]
        self.cache_boxes = np.zeros((MAX_FRAMES, 15), dtype=np.float32) 
        
        self.cache_x_out = None
        self.cache_sim_mat = None
        self.cache_sim_score = None

        self.current_id = 0
        self.last_update_id = -99


    def get_similarity(self, x_out, proposals, meta_info=None):
        '''
        Args:
            x_feats: list of torch.Tensor
            proposal: numpy array
        '''

        with torch.no_grad():
            num_proposals = proposals.shape[0]
            x_rois_np = np.hstack((np.zeros((num_proposals, 1)), proposals))
            x_rois = torch.from_numpy(x_rois_np).float().to(self.model_device)
            sim_cls_logit, sim_box_deltas = \
                self.extract_sim_mat(x_out=x_out, x_rois=x_rois)
            # sim_cls_logit [Num_templates, num_rois, 2]
            # sim_box_deltas [Num templates, num_rois, 4]
            sim_prob = F.softmax(sim_cls_logit, dim=2) # [Num_templates, num_rois, 2]
        
        num_pos_temp = self.num_temps
        temp_weight_list_pos = self.temp_weights[self.temp_inds].reshape(num_pos_temp, 1, 1)
        sim_score_all = sim_prob.cpu().numpy()
        bbox_delta_all = sim_box_deltas.data.cpu().numpy()

        # sum
        sim_score_list_pos = sim_score_all[:num_pos_temp, :, :]
        bbox_delta_list_pos = bbox_delta_all[:num_pos_temp, :, :]

        sim_score_pos = (sim_score_list_pos * temp_weight_list_pos).sum(axis=0) / temp_weight_list_pos.sum()
        bbox_weights = sim_score_list_pos[:, :, [1]] * temp_weight_list_pos
        bbox_delta_pos = (bbox_delta_list_pos * bbox_weights).sum(axis=0) / bbox_weights.sum(axis=0)

        sim_score = sim_score_pos[:, 1]
        bbox_pred = ubox.bbox_transform(proposals[:, :4], bbox_delta_pos, weights=cfg.RCNN.BBOX_VAR_INV)
        
        if self.use_negative:
            num_neg_temp = self.num_neg_temps

            if num_neg_temp > 0:    
                # sum
                sim_score_list_neg = sim_score_all[num_pos_temp:, :, :]
                temp_weight_list_neg = self.neg_temp_weights[self.neg_temp_inds].reshape(num_neg_temp, 1, 1)

                sim_score_neg = (sim_score_list_neg * temp_weight_list_neg).sum(axis=0) / temp_weight_list_neg.sum()
                sim_score = sim_score - sim_score_neg[:, 1] * cfg.INFERENCE.NEGATIVE_WEIGHTS


        # print('----------')
        # for k in range(sim_score_all.shape[0]):
        #     if k < num_pos_temp:
        #         print('Positive {} {:.2f}'.format(k, self.temp_weights[self.temp_inds[k]]), end='|')
        #     else:
        #         print('Negative {} {:.2f}'.format(k, self.neg_temp_weights[self.neg_temp_inds[k-num_pos_temp]]), end='|')
        #     for c in range(sim_score_all.shape[1]):
        #         print('{:.2f} '.format(sim_score_all[k, c, 1]), end='')
        #     print('')
        # print('-----------')

        
        if isinstance(meta_info, dict):
            meta_info['sim_mat_pos'] = sim_score_all[:num_pos_temp, :, 1].copy()
            
            meta_info['sim_weight_pos'] = self.temp_weights[self.temp_inds].copy()
            if self.use_negative:
                meta_info['sim_mat_neg'] = sim_score_all[num_pos_temp:, :, 1].copy()
                meta_info['sim_weight_neg'] = self.temp_weights[self.neg_temp_inds].copy()

        self.cache_sim_score = sim_score # [K, ]
        self.cache_sim_mat = sim_score_all[:, :, 1].copy() # [N, K]
        self.cache_x_out = x_out 
        return sim_score, bbox_pred
        

    def init_template(self, z_feats):
        self.add_template(z_feats, 1.0, is_positive=True, enable_update=False)


    def init_module(self):
        self.temp_weights = np.zeros((self.max_num_temps, ), dtype=np.float32)
        self.temp_inds = []
        self.temp_enable_update = np.zeros((self.max_num_temps, ), dtype=np.bool)
        self.num_temps = 0

        if self.use_negative:
            self.neg_temp_weights = np.zeros((self.max_num_neg_temps, ), dtype=np.float32)
            self.neg_temp_inds = []
            self.neg_temp_enable_update = np.zeros((self.max_num_neg_temps, ), dtype=np.bool)
            self.num_neg_temps = 0
            
        self.cache_boxes = np.zeros((MAX_FRAMES, 15), dtype=np.float32) 
        
        self.cache_x_out = None
        self.cache_sim_mat = None
        self.cache_sim_score = None

        self.current_id = 0
        self.last_update_id = -99

    
    def add_template(self, z_feats, weights, is_positive=True, enable_update=True):

        if is_positive:
            if self.num_temps >= self.max_num_temps:
                raise ValueError('Achieve the max number of templates')
            next_idx_set = set(range(self.max_num_temps)) - set(self.temp_inds)
            next_idx = next_idx_set.pop()
            self.temp_embeddings[next_idx, :, :, :] = z_feats.view(self.z_channels, self.z_size, self.z_size)
            self.temp_weights[next_idx] = weights
            self.temp_enable_update[next_idx] = enable_update
            self.temp_inds.append(next_idx)
            self.num_temps += 1      
        
        else:
            assert self.use_negative
            if self.num_neg_temps >= self.max_num_neg_temps:
                raise ValueError('Achieve the max number of templates')
            next_idx_set = set(range(self.max_num_neg_temps)) - set(self.neg_temp_inds)
            next_idx = next_idx_set.pop()
            self.neg_temp_embeddings[next_idx, :, :, :] = z_feats.view(self.z_channels, self.z_size, self.z_size)
            self.neg_temp_weights[next_idx] = weights
            self.neg_temp_enable_update[next_idx] = enable_update
            self.neg_temp_inds.append(next_idx)
            self.num_neg_temps += 1
        

    def remove_template(self, temp_idx, is_positive=True):
        if is_positive:
            del self.temp_inds[temp_idx]
            self.num_temps -= 1
        else:
            del self.neg_temp_inds[temp_idx]
            self.num_neg_temps -= 1


    def extract_sim_mat(self, x_out, x_rois):
        x_feats = self.model.feat_extractor(x_out, x_rois)

        num_x_rois = x_rois.size(0)
        num_temps = len(self.temp_inds)
            
        sim_mat = []
        bbox_mat = []

        for i in self.temp_inds:
            fusion_feat = self.model.siam_fusion(self.temp_embeddings[i].unsqueeze(0), x_feats)
            cls_logit = self.model.rcnn_cls(fusion_feat) # N x 2
            bbox_delta = self.model.rcnn_bbox(fusion_feat)
            sim_mat.append(cls_logit.unsqueeze(0))
            bbox_mat.append(bbox_delta.unsqueeze(0))

        if self.use_negative:
            for i in self.neg_temp_inds:
                fusion_feat = self.model.siam_fusion(self.neg_temp_embeddings[i].unsqueeze(0), x_feats)
                cls_logit = self.model.rcnn_cls(fusion_feat) # N x 2
                bbox_delta = self.model.rcnn_bbox(fusion_feat)
                sim_mat.append(cls_logit.unsqueeze(0))
                bbox_mat.append(bbox_delta.unsqueeze(0))

        sim_mat = torch.cat(sim_mat, dim=0)
        bbox_mat = torch.cat(bbox_mat, dim=0)

        return sim_mat, bbox_mat


    def update(self, box_in_img, box_in_sr, rpn_score, win_score, max_score_idx, meta_info=None):

        idx = self.current_id
        # save information to caches
        num_pos_temp = len(self.temp_inds)

        self.cache_boxes[idx, 0:4] = box_in_img[max_score_idx]
        self.cache_boxes[idx, 4:8] = box_in_sr[max_score_idx]
        self.cache_boxes[idx, 8] = rpn_score[max_score_idx]
        self.cache_boxes[idx, 9] = win_score[max_score_idx]
        self.cache_boxes[idx, 10] = self.cache_sim_mat[:num_pos_temp, max_score_idx].max()
        self.cache_boxes[idx, 11] = self.cache_sim_score[max_score_idx]

        if idx == 0:
            size_penalty = 0.
            shape_penalty = 0.
        else:
            lw = self.cache_boxes[idx-1, 2] - self.cache_boxes[idx-1, 0]
            lh = self.cache_boxes[idx-1, 3] - self.cache_boxes[idx-1, 1]
            cw = self.cache_boxes[idx, 2] - self.cache_boxes[idx, 0]
            ch = self.cache_boxes[idx, 3] - self.cache_boxes[idx, 1]

            last_size = lw * lh
            current_size = cw * ch
            size_penalty = np.abs(np.log(current_size / last_size))

            last_ar = lw / lh
            current_ar = cw / ch
            shape_penalty = np.abs(np.log(current_ar / last_ar))
        self.cache_boxes[idx, 12] = size_penalty
        self.cache_boxes[idx, 13] = shape_penalty
            


        # [x, y, h, w, rpn_score, win_score, max_sim_score, avg_sim_score]

        if self.update_rule == 0:
            update_info = self._naive_update(box_in_sr, rpn_score, max_score_idx)
        if self.update_rule == 1:
            update_info = self._pos_only_update(box_in_sr, rpn_score, max_score_idx)
        self.current_id += 1

        if isinstance(meta_info, dict):
            meta_info['update_pos'] = update_info[0]
            if self.use_negative:
                meta_info['update_neg'] = update_info[1]
    

    def _pos_only_update(self, box_in_sr, rpn_score, max_score_idx):

        idx = self.current_id
        check_idx_start = max(0, idx - cfg.INFERENCE.UPDATE_PO_CHECK_STEP)
        
        # update weights of templates
        _w = cfg.INFERENCE.UPDATE_PO_RUNNING_MEAN_WEIGHT
        updated_weights = np.zeros((self.max_num_temps, ), dtype=np.float32)
        updated_weights[self.temp_inds] = self.temp_weights[self.temp_inds] * _w + \
                          self.cache_sim_mat[:self.num_temps, max_score_idx] * (1 - _w) * cfg.INFERENCE.UPDATE_PO_WEIGHT_FACTOR
        self.temp_weights[self.temp_enable_update] = updated_weights[self.temp_enable_update]

        if idx < self.last_update_id + cfg.INFERENCE.UPDATE_PO_MIN_STEP:
            return [(), ()]

        # conditions:
        # 1) RPN score 
        rpn_score_cond = self.cache_boxes[idx, 8] >= cfg.INFERENCE.UPDATE_PO_RPN_SCORE_MIN
        # 2) Similarity score
        sim_score_cond = (self.cache_boxes[idx, 10] < cfg.INFERENCE.UPDATE_PO_SIM_SCORE_MAX) and \
                         (self.cache_boxes[idx, 10] >= cfg.INFERENCE.UPDATE_PO_SIM_SCORE_MIN)
        # 3) stable condition
        win_score_cond = (self.cache_boxes[check_idx_start:idx+1, 9].min() > cfg.INFERENCE.UPDATE_PO_WIN_SCORE_MIN)
        size_penalty_cond = (self.cache_boxes[check_idx_start:idx+1, 12].max() < cfg.INFERENCE.UPDATE_PO_SIZE_PENALTY_MAX)
        ar_penalty_cond = (self.cache_boxes[check_idx_start:idx+1, 13].max() < cfg.INFERENCE.UPDATE_PO_AR_PENALTY_MAX)

        update_cond = rpn_score_cond and sim_score_cond and win_score_cond and size_penalty_cond and ar_penalty_cond

        # print('RPN score   : {:.2f} {}'.format(self.cache_boxes[idx, 8], rpn_score_cond))
        # print('Sim score   : {:.2f} {}'.format(self.cache_boxes[idx, 10], sim_score_cond))
        # print('Win score   : {:.2f} {}'.format(self.cache_boxes[check_idx_start:idx+1, 9].min(), win_score_cond))
        # print('Size penalty: {:.2f} {}'.format(self.cache_boxes[check_idx_start:idx+1, 12].max(), size_penalty_cond))
        # print('AR penalty  : {:.2f} {}'.format(self.cache_boxes[check_idx_start:idx+1, 13].max(), ar_penalty_cond))

        if update_cond:
            if self.num_temps >= self.max_num_temps:
                # drop out the templates with smallest weights
                smallest_temp_idx = np.argmin(self.temp_weights[self.temp_inds])
                self.remove_template(smallest_temp_idx, is_positive=True)
            x_rois_np = np.zeros((1, 5), dtype=np.float32)
            x_rois_np[0, 1:5] = self.cache_boxes[idx, 4:8]
            x_rois = torch.from_numpy(x_rois_np).type_as(self.temp_embeddings)
            x_feats = self.model.feat_extractor(self.cache_x_out, x_rois)

            self.add_template(x_feats, self.cache_boxes[idx, 8] * cfg.INFERENCE.UPDATE_PO_WEIGHT_FACTOR, is_positive=True)

            self.last_update_id = idx

            return [(max_score_idx), ()]
        return [(), ()]


    def _naive_update(self, box_in_sr, rpn_score, max_score_idx, debug_info=True):
        idx = self.current_id
        # win_idx_s = max(0, idx - cfg.INFERENCE.UPDATE_NAIVE_WIN_SCORE_STEP)
        
        num_pos_temp = len(self.temp_inds)
         # update weights of templates (positive)
        _w = cfg.INFERENCE.UPDATE_NAIVE_RUNNING_MEAN_WEIGHT
        updated_weights = np.zeros((self.max_num_temps, ), dtype=np.float32)
        updated_weights[self.temp_inds] = self.temp_weights[self.temp_inds] * _w + \
                          self.cache_sim_mat[:num_pos_temp, max_score_idx] * (1 - _w)
        self.temp_weights[self.temp_enable_update] = updated_weights[self.temp_enable_update]

        # update weights of tempaltes (negative)
        if self.use_negative:
            if self.num_neg_temps > 0:
                updated_weights = np.zeros((self.max_num_neg_temps, ), dtype=np.float32)
                updated_weights[self.neg_temp_inds] = self.neg_temp_weights[self.neg_temp_inds] * _w + \
                                self.cache_sim_mat[num_pos_temp:, max_score_idx] * (1 - _w)
                self.neg_temp_weights[self.neg_temp_enable_update] = updated_weights[self.neg_temp_enable_update]
        
        # Conditions of updating positive sample
        # 1) RPN score mush higher than a thresh (cfg.INFERENCE.UPDATE_NAIVE_RPN_SCORE_TH)
        # 2) The simiarity scores to the existing positive templates mush lower than a thresh
        # 3) (if use negative) The simiarity scores to the existing negative template mush lower than a thresh

        rpn_score_cond = self.cache_boxes[idx, 8] >= cfg.INFERENCE.UPDATE_NAIVE_POS_RPN_SCORE_TH
        sim_score_cond = self.cache_boxes[idx, 10] < self.cache_boxes[idx, 8] - cfg.INFERENCE.UPDATE_NAIVE_POS_SIMPOS_SCORE_TH
        # sim_score_cond = self.cache_boxes[idx, 10] < cfg.INFERENCE.UPDATE_NAIVE_SIM_SCORE_TH
        # win_score_cond = self.cache_boxes[win_idx_s:idx+1, 9].min() >= cfg.INFERENCE.UPDATE_NAIVE_WIN_SCORE_TH
        neg_score_cond = True
        if self.use_negative and self.num_neg_temps > 0:
            neg_score_cond = self.cache_sim_mat[num_pos_temp:, max_score_idx].max() < \
                            cfg.INFERENCE.UPDATE_NAIVE_POS_SIMNGE_SCORE_TH
        update_positive = rpn_score_cond and \
                          sim_score_cond and \
                          neg_score_cond

        if update_positive:

            if self.num_temps >= self.max_num_temps:
                # drop out the templates with smallest weights
                smallest_temp_idx = np.argmin(self.temp_weights[self.temp_inds])
                self.remove_template(smallest_temp_idx, is_positive=True)
            x_rois_np = np.zeros((1, 5), dtype=np.float32)
            x_rois_np[0, 1:5] = self.cache_boxes[idx, 4:8]
            x_rois = torch.from_numpy(x_rois_np).type_as(self.temp_embeddings)
            x_feats = self.model.feat_extractor(self.cache_x_out, x_rois)

            self.add_template(x_feats, self.cache_boxes[idx, 8], is_positive=True)
            if debug_info:
                print('Update positive template!')

        if self.use_negative:
            # Conditions of updating negative sample
            # 1) RPN score mush higher than a thresh (cfg.INFERENCE.UPDATE_NAIVE_NEG_RPN_SCORE_TH)
            # 2) Overlaps with the selected object must lower than a threshold
            # 3) The simiarity scores to the existing negative templates mush lower than a thresh
            
            if self.num_neg_temps > 0:
                max_neg_sim = self.cache_sim_mat[num_pos_temp:, :].max(axis=0)
            else:
                max_neg_sim = np.zeros((self.cache_sim_mat.shape[1], ), dtype=np.float32)

            overlaps = ubox.bbox_overlaps(np.ascontiguousarray(box_in_sr[:, :4], dtype=np.float32), 
                    box_in_sr[max_score_idx, :4].reshape(1, 4)).flatten()
            
            update_negative_inds = np.where((rpn_score >= cfg.INFERENCE.UPDATE_NAIVE_NEG_RPN_SCORE_TH) &
                                            (max_neg_sim < cfg.INFERENCE.UPDATE_NAIVE_NEG_SIMNEG_SCORE_TH) &
                                            (overlaps <cfg.INFERENCE.UPDATE_NAIVE_NEG_OVERLAP_TH))[0]
            if debug_info:
                #print('[Negative update conds]')
                #print((rpn_score >= cfg.INFERENCE.UPDATE_NAIVE_NEG_RPN_SCORE_TH))
                #print((max_neg_sim < cfg.INFERENCE.UPDATE_NAIVE_NEG_SIMNEG_SCORE_TH))
                #print(overlaps <cfg.INFERENCE.UPDATE_NAIVE_NEG_OVERLAP_TH)
                pass

            num_updates = len(update_negative_inds)
            if num_updates > 0:
                update_negative_inds = update_negative_inds[np.argsort(rpn_score[update_negative_inds])[::-1]]

                x_rois_np = np.zeros((num_updates, 5), dtype=np.float32)
                x_rois_np[:, 1:5] = box_in_sr[update_negative_inds, :4]
                x_rois = torch.from_numpy(x_rois_np).type_as(self.neg_temp_embeddings)
                x_feats = self.model.feat_extractor(self.cache_x_out, x_rois)

                merged = np.zeros((num_updates, ), np.bool)
                for i, neg_idx in enumerate(update_negative_inds):
                    if merged[i]:
                        continue
                    if self.num_neg_temps >= self.max_num_neg_temps:
                        # drop out the templates with smallest weights
                        smallest_temp_idx = np.argmin(self.neg_temp_weights[self.neg_temp_inds])
                        self.remove_template(smallest_temp_idx, is_positive=False)
                    self.add_template(x_feats[i], rpn_score[neg_idx], is_positive=False)

                    if debug_info:
                        print('Add negative template')

                    fusion_feat = self.model.siam_fusion(x_feats[i].unsqueeze(0), x_feats)
                    cls_logit = self.model.rcnn_cls(fusion_feat) # N x 2
                    sim_prob = F.softmax(cls_logit, dim=1).data.cpu().numpy()

                    merge_inds = np.where(sim_prob[:, 1] >= cfg.INFERENCE.UPDATE_NAIVE_NEG_SIMNEG_SCORE_TH)[0]
                    merged[merge_inds] = True
