import torch
import math
from torch import nn
from collections import OrderedDict


class AlexNetConv(nn.Module):

    def __init__(self, padding=False, all_layer_out=False):
        super(AlexNetConv, self).__init__()
        self.all_layer_out = all_layer_out
        if not padding:
            self.conv1 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(3, 96, 11, stride=2, padding=0, bias=True)),
                ('relu', nn.ReLU()),
                ('pool', nn.MaxPool2d(3, stride=2, padding=0, dilation=1)),
                ('norm', nn.LocalResponseNorm(size=5, alpha=1e-4, beta=0.75, k=1.0))]))

            self.conv2 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(96, 256, 5, stride=1, padding=0, groups=2, bias=True)),
                ('relu', nn.ReLU()),
                ('pool', nn.MaxPool2d(3, stride=2, padding=0, dilation=1)),
                ('norm', nn.LocalResponseNorm(size=5, alpha=1e-4, beta=0.75, k=1.0))]))

            self.conv3 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(256, 384, kernel_size=3, padding=0, bias=True)),
                ('relu', nn.ReLU())]))

            self.conv4 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(384, 384, kernel_size=3, padding=0, groups=2, bias=True)),
                ('relu', nn.ReLU())]))

            self.conv5 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(384, 256, kernel_size=3, padding=0, groups=2, bias=True)),
                ('relu', nn.ReLU())]))
        else:
            self.conv1 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(3, 96, 11, stride=2, padding=5, bias=True)),
                ('relu', nn.ReLU()),
                ('pool', nn.MaxPool2d(3, stride=2, padding=1, dilation=1)),
                ('norm', nn.LocalResponseNorm(size=5, alpha=1e-4, beta=0.75, k=1.0))]))

            self.conv2 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(96, 256, 5, stride=1, padding=2, groups=2, bias=True)),
                ('relu', nn.ReLU()),
                ('pool', nn.MaxPool2d(3, stride=2, padding=1, dilation=1)),
                ('norm', nn.LocalResponseNorm(size=5, alpha=1e-4, beta=0.75, k=1.0))]))

            self.conv3 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(256, 384, kernel_size=3, padding=1, bias=True)),
                ('relu', nn.ReLU())]))

            self.conv4 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(384, 384, kernel_size=3, padding=1, groups=2, bias=True)),
                ('relu', nn.ReLU())]))

            self.conv5 = nn.Sequential(OrderedDict([
                ('conv', nn.Conv2d(384, 256, kernel_size=3, padding=1, groups=2, bias=True)),
                ('relu', nn.ReLU())]))
        self._initialize_weights()

    def forward(self, x):
        if not self.all_layer_out:
            out = self.conv1(x)
            out = self.conv2(out)
            out = self.conv3(out)
            out = self.conv4(out)
            out = self.conv5(out)
        else:
            out = {}
            out['conv1'] = self.conv1(x)
            out['conv2'] = self.conv2(out['conv1'])
            out['conv3'] = self.conv3(out['conv2'])
            out['conv4'] = self.conv4(out['conv3'])
            out['conv5'] = self.conv5(out['conv4'])
        return out

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                output_channel = m.weight.shape[0]
                num_groups = m.groups
                channels_per_group = int(output_channel / num_groups)
                for i in range(num_groups):
                    st = channels_per_group * i
                    end = channels_per_group * (i+1)
                    torch.nn.init.xavier_uniform_(m.weight[st:end, :, :, :])
                if m.bias is not None:
                    m.bias.data.zero_()

    def freeze_conv_layer(self, num_conv = 5):
        for i in range(1, num_conv + 1):
            for module_name, m in self.named_modules():
                if module_name.startswith('conv{}'.format(i)):
                    if isinstance(m, torch.nn.Conv2d):
                        m.weight.requires_grad = False
                        if m.bias is not None:
                            m.bias.requires_grad = False
                        print("Freeze conv layer {}".format(module_name))
                    if isinstance(m, torch.nn.BatchNorm2d):
                        if m.weight is not None:
                            m.weight.requires_grad = False
                        if m.bias is not None:
                            m.bias.requires_grad = False
                        m.eval()
                        print("Freeze bn layer {}".format(module_name))


class AlexNetConvPlus512(nn.Module):

    def __init__(self):
        super(AlexNetConvPlus512, self).__init__()

        self.conv1 = nn.Sequential(OrderedDict([
            ('conv', nn.Conv2d(3, 96, 11, stride=2, padding=0, bias=True)),
            ('relu', nn.ReLU()),
            ('pool', nn.MaxPool2d(3, stride=2, padding=0, dilation=1)),
            ('norm', nn.LocalResponseNorm(size=5, alpha=1e-4, beta=0.75, k=1.0))]))

        self.conv2 = nn.Sequential(OrderedDict([
            ('conv', nn.Conv2d(96, 256, 5, stride=1, padding=0, groups=2, bias=True)),
            ('relu', nn.ReLU()),
            ('pool', nn.MaxPool2d(3, stride=2, padding=0, dilation=1)),
            ('norm', nn.LocalResponseNorm(size=5, alpha=1e-4, beta=0.75, k=1.0))]))

        self.conv3 = nn.Sequential(OrderedDict([
            ('conv', nn.Conv2d(256, 384, kernel_size=3, padding=0, bias=True)),
            ('relu', nn.ReLU())]))

        self.conv4 = nn.Sequential(OrderedDict([
            ('conv', nn.Conv2d(384, 384, kernel_size=3, padding=0, groups=2, bias=True)),
            ('relu', nn.ReLU())]))

        self.conv5 = nn.Sequential(OrderedDict([
            ('conv', nn.Conv2d(384, 256, kernel_size=3, padding=0, groups=2, bias=True)),
            ('relu', nn.ReLU())]))
        
        self.conv6 = nn.Sequential(OrderedDict([
            ('conv', nn.Conv2d(256, 512, kernel_size=1, padding=0, groups=1, bias=True)),
            ('relu', nn.ReLU())]))

        self._initialize_weights()

    def forward(self, x):

        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)
        out = self.conv4(out)
        out = self.conv5(out)
        out = self.conv6(out)
        return out

    def _initialize_weights(self):
        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                output_channel = m.weight.shape[0]
                num_groups = m.groups
                channels_per_group = int(output_channel / num_groups)
                for i in range(num_groups):
                    st = channels_per_group * i
                    end = channels_per_group * (i+1)
                    torch.nn.init.xavier_uniform_(m.weight[st:end, :, :, :])
                if m.bias is not None:
                    m.bias.data.zero_()

    def freeze_conv_layer(self, num_conv = 5):
        for i in range(1, num_conv + 1):
            for module_name, m in self.named_modules():
                if module_name.startswith('conv{}'.format(i)):
                    if isinstance(m, torch.nn.Conv2d):
                        m.weight.requires_grad = False
                        if m.bias is not None:
                            m.bias.requires_grad = False
                        print("Freeze conv layer {}".format(module_name))
                    if isinstance(m, torch.nn.BatchNorm2d):
                        if m.weight is not None:
                            m.weight.requires_grad = False
                        if m.bias is not None:
                            m.bias.requires_grad = False
                        m.eval()
                        print("Freeze bn layer {}".format(module_name))


if __name__ == '__main__':
    net_pad = AlexNetConv(padding=True, all_layer_out=True)
    net_wo_pad = AlexNetConv(padding=False, all_layer_out=True)
    x = torch.zeros(1, 3, 239, 239)
    out_pad = net_pad(x)
    out_wo_pad = net_wo_pad(x)
    for k in range(1, 6):
        str_key = "conv{}".format(k)
        print('{} ---> {}, {}'.format(str_key, out_pad[str_key].size(), out_wo_pad[str_key].size()))