import torch
import torch.nn as nn
import numpy as np
import math
import yaml

from core.config import cfg
from utils.nms import nms_torch
from rpn.generate_anchors import generate_anchors
from utils.boxes import bbox_transform_batch, bbox_transform_inv

import time

class _ProposalLayer(nn.Module):
    """
    Outputs object detection proposals by applying estimated bounding-box
    transformations to a set of regular boxes (called "anchors").
    """

    def __init__(self, feat_stride, scales, ratios, is_training=True, anchor_shift=(63., 63.), force_cpu=False):
        super(_ProposalLayer, self).__init__()

        self._feat_stride = feat_stride
        self._anchors = torch.from_numpy(generate_anchors(scales=np.array(scales), 
            ratios=np.array(ratios))).float()
        self._anchors[:, 0::2] += anchor_shift[0]
        self._anchors[:, 1::2] += anchor_shift[0] 
        self._num_anchors = self._anchors.size(0)

        if cfg.SIAMRPN.RPN_HEAD_LOSS == 'sigmoid':
            self._score_start_idx = 0
        elif cfg.SIAMRPN.RPN_HEAD_LOSS == 'softmax':
            self._score_start_idx = self._num_anchors

        self.is_training = is_training
        if self.is_training:
            self.cfg_key = 'TRAIN'
        else:
            self.cfg_key = 'INFERENCE'
        self.force_cpu = force_cpu

    def forward(self, in_data):
        if self.force_cpu:
            scores = in_data[0][:, self._score_start_idx:, :, :].cpu().contiguous() # N x A x H x W
            bbox_deltas = in_data[1].cpu() # N x 4A x H x W
        else:
            scores = in_data[0][:, self._score_start_idx:, :, :].contiguous() # N x A x H x W
            bbox_deltas = in_data[1] # N x 4A x H x W

        post_nms_topN = cfg[self.cfg_key].RPN_POST_NMS_TOP_N
        nms_thresh    = cfg[self.cfg_key].RPN_NMS_THRESH
        # min_size      = cfg[self.cfg_key].RPN_MIN_SIZE

        batch_size = bbox_deltas.size(0)
        feat_height, feat_width = scores.size(2), scores.size(3)

        shift_x = torch.arange(0, feat_width) * self._feat_stride
        shift_y = torch.arange(0, feat_height) * self._feat_stride
        yv, xv = torch.meshgrid([shift_x, shift_y])
        yv = yv.contiguous().view(-1, 1)
        xv = xv.contiguous().view(-1, 1)
        shifts = torch.cat((xv, yv, xv, yv), dim=1)
        shifts = shifts.contiguous().type_as(scores).float()

        A = self._num_anchors
        K = shifts.size(0)
        self._anchors = self._anchors.type_as(scores)

        anchors = self._anchors.view(1, A, 4) + shifts.view(K, 1, 4)
        anchors = anchors.view(1, K * A, 4).expand(batch_size, K * A, 4)

        # Transpose and reshape predicted bbox transformations to get them
        # into the same order as the anchors:
        bbox_deltas = bbox_deltas.permute(0, 2, 3, 1).contiguous()
        bbox_deltas = bbox_deltas.view(batch_size, -1, 4)

        # Same story for the scores:
        scores = scores.permute(0, 2, 3, 1).contiguous()
        scores = scores.view(batch_size, -1)

        # Convert anchors into proposals via bbox transformations
        proposals = bbox_transform_batch(anchors, bbox_deltas, batch_size)

        # sort for NMS
        scores_keep = scores
        proposals_keep = proposals
        _, order = torch.sort(scores_keep, 1, True)

        output = scores.new(batch_size, post_nms_topN, 5).zero_() - 1
        for i in range(batch_size):
            # # 3. remove predicted boxes with either height or width < threshold
            proposals_single = proposals_keep[i]
            scores_single = scores_keep[i]

            # # 4. sort all (proposal, score) pairs by score from highest to lowest
            # # 5. take top pre_nms_topN (e.g. 6000)
            order_single = order[i]

            proposals_single = proposals_single[order_single, :]
            scores_single = scores_single[order_single].view(-1,1)

            # 6. apply nms (e.g. threshold = 0.7)
            # 7. take after_nms_topN (e.g. 300)
            # 8. return the top proposals (-> RoIs top)
            
            keep_idx_i = nms_torch(torch.cat((proposals_single, scores_single), 1), nms_thresh)
            keep_idx_i = keep_idx_i.long().view(-1)

            if post_nms_topN > 0:
                keep_idx_i = keep_idx_i[:post_nms_topN]
            proposals_single = proposals_single[keep_idx_i, :]
            scores_single = scores_single[keep_idx_i, :]

            # padding 0 at the end.
            num_proposal = proposals_single.size(0)
            output[i,:,0] = i
            output[i,:num_proposal,1:] = proposals_single

        if self.force_cpu:
            output = output.type_as(in_data[1])
        return output


    def _filter_boxes(self, boxes, min_size):
        """Remove all boxes with any side smaller than min_size."""
        ws = boxes[:, :, 2] - boxes[:, :, 0]
        hs = boxes[:, :, 3] - boxes[:, :, 1]
        keep = ((ws >= min_size.view(-1,1).expand_as(ws)) & (hs >= min_size.view(-1,1).expand_as(hs)))
        return keep
