import torch
import torch.nn.functional as F
import numpy as np
from copy import deepcopy
from collections import OrderedDict

from models.base_conv import model_factory
from utils.model_weights import random_init_weights
from utils.crop_and_resize import CropAndResizeFunction

from utils.misc import xcorr

class _Flatten(torch.nn.Module):
    def forward(self, x):
        return x.view(x.size(0), -1)


class FeatExtractor(torch.nn.Module):

    def __init__(self, cfg, is_training=False):
        super(FeatExtractor, self).__init__()

        self.cfg = cfg

        fusion_type_dict = {
            'concat': 0,
            'sum': 1
        }

        
        self.out_height, self.out_width = cfg.RCNN.POOL_SIZE
        self.out_channel = cfg.RCNN.POOL_CHANNEL

        self.use_multi_layer = len(cfg.RCNN.FEAT_LIST) > 1
        if self.use_multi_layer:
            self.feat_name_list = cfg.RCNN.FEAT_LIST
            self.stride_list = cfg.RCNN.FEAT_STRIDE_LIST
            self.fusion_type = fusion_type_dict[cfg.RCNN.MULTI_LAYER_FUSION_TYPE]

            if self.fusion_type == 1:
                # 1x1 conv
                self.conv1x1_list = torch.nn.ModuleList(
                    [torch.nn.Conv2d(cfg.RCNN.FEAT_CHANNELS[i], self.out_channel, 1) for i in range(len(self.feat_name_list))]
                )
            self.use_bn = cfg.RCNN.ROI_FEAT_BN
            if self.use_bn:
                self.bn_list = torch.nn.ModuleList(
                    [torch.nn.BatchNorm2d(cfg.RCNN.FEAT_CHANNELS[i]) for i in range(len(self.feat_name_list))]
                )

            self.feat_eq_size = True
            for _s in self.stride_list:
                if _s != self.stride_list[0]:
                    self.feat_eq_size = False
            
            
        else:
            self.feat_name = cfg.RCNN.FEAT_LIST[0]
            self.stride = cfg.RCNN.FEAT_STRIDE_LIST[0]

    def forward(self, feats, rois):
        if not self.use_multi_layer:
            num_imgs, channels, height, width = feats[self.feat_name].size()
        else:
            num_imgs, channels, height, width = feats[self.feat_name_list[0]].size()
    
        roi_boxes = rois[:, 1:5].contiguous()
        roi_inds = rois[:, 0].contiguous()
        
        if self.use_multi_layer:
            if self.feat_eq_size:
                # currently only support stride_list to be a scalar
                roi_boxes = roi_boxes / self.stride_list[0]
                if self.fusion_type == 0:
                    # concat all the feature
                    if self.use_bn:
                        concat_feat = torch.cat([self.bn_list[ix](feats[name])
                                for ix, name in enumerate(self.feat_name_list)], dim=1).contiguous()
                    else:
                        concat_feat = torch.cat([feats[name] for name in self.feat_name_list], dim=1).contiguous()
                    roi_feats = CropAndResizeFunction(self.out_height, self.out_width, has_normed=False)(
                                    concat_feat, roi_boxes, roi_inds)
                elif self.fusion_type == 1:
                    if self.use_bn:
                        sum_feat = torch.sum(torch.cat([self.bn_list[ix](self.conv1x1_list[ix](feats[name])).unsqueeze(0) 
                            for ix, name in enumerate(self.feat_name_list)], dim=0), dim=0)
                    else:
                        sum_feat = torch.sum(torch.cat([self.conv1x1_list[ix](feats[name]).unsqueeze(0) 
                            for ix, name in enumerate(self.feat_name_list)], dim=0), dim=0)
                    roi_feats = CropAndResizeFunction(self.out_height, self.out_width, has_normed=False)(
                                    sum_feat, roi_boxes, roi_inds)
                else:
                    raise ValueError('Unknown type {}'.format(self.fusion_type))
            else:
                if self.fusion_type == 1:
                    conv_feats = {name: self.conv1x1_list[ix](feats[name])
                                        for ix, name in enumerate(self.feat_name_list)}
                else:
                    conv_feats = {name: feats[name]
                                        for ix, name in enumerate(self.feat_name_list)}

                if self.use_bn:
                    bn_feats = {name: self.bn_list[ix](conv_feats[name])
                                for ix, name in enumerate(self.feat_name_list)}
                else:
                    bn_feats = {name: conv_feats[name] 
                                for ix, name in enumerate(self.feat_name_list)}
                

                crop_feats = []
                for ix, name in enumerate(self.feat_name_list):
                    i_roi_boxes = roi_boxes / self.stride_list[ix]
                    crop_feats.append(
                        CropAndResizeFunction(self.out_height, self.out_width, has_normed=False)(
                            bn_feats[name], i_roi_boxes, roi_inds
                        )
                    )
                if self.fusion_type == 0:
                    roi_feats = torch.cat(crop_feats, dim=1).contiguous()
                
                elif self.fusion_type == 1:
                    for ix, name in enumerate(self.feat_name_list):
                        crop_feats[ix] = crop_feats[ix].unsqueeze(0)
                    roi_feats = torch.sum(crop_feats, dim=0).contiguous()

        else:
            # feats is a Tensor in shape of [N x C x H x W]
            # rois is a Tensor in shape of [K x 5] or [N x 4]
            # normalize roi_boxes to feature domain
            
            roi_boxes /= self.stride
            roi_feats = CropAndResizeFunction(self.out_height, self.out_width, has_normed=False)(
                            feats[self.feat_name], roi_boxes, roi_inds)
            
        return roi_feats


class SiamFeatFusion(torch.nn.Module):

    def __init__(self, cfg, is_training=False):
        super(SiamFeatFusion, self).__init__()

        fusion_type_dict = {
            'concat': 0,
            'diff': 1,
            'flat_concat': 2,
            'flat_diff': 3,
            'correlation': 4,
        }

        self.fusion_type = fusion_type_dict[cfg.RCNN.SIAM_FEAT_FUSION_TYPE]
        
        if self.fusion_type != 4:
            force_linear = cfg.RCNN.HIDDEN_LAYER_FORCE_LINEAR
            self.in_height, self.in_width = cfg.RCNN.POOL_SIZE
            self.in_channel = cfg.RCNN.POOL_CHANNEL

            if self.fusion_type == 0:
                self.in_channel *= 2

            self.num_conv = len(cfg.RCNN.POOL_CONVS)
            if self.num_conv > 0:
                self.roi_conv = torch.nn.Sequential()
                for i in range(self.num_conv):
                    c = cfg.RCNN.POOL_CONVS[i]
                    self.roi_conv.add_module('roi_conv{}'.format(i), 
                        torch.nn.Conv2d(self.in_channel, c, 3, stride=1, padding=1))
                    self.in_channel = c
                    if not force_linear:
                        self.roi_conv.add_module('roi_relu{}'.format(i),
                            torch.nn.ReLU(inplace=True))

            self.hidden_channel = cfg.RCNN.HIDDEN_LAYER_CHANNEL

            self.out_channel = self.in_channel * self.in_height * self.in_width
            self.num_linear = cfg.RCNN.NUM_LINEAR_LAYERS

            if self.fusion_type in (2, 3):
                raise NotImplementedError # disable
                if force_linear:
                    self.pre_linear = torch.nn.Sequential(
                                torch.nn.Linear(self.out_channel,self.hidden_channel))
                else:
                    self.pre_linear = torch.nn.Sequential(
                                torch.nn.Linear(self.out_channel,self.hidden_channel),
                                torch.nn.ReLU())
                if self.fusion_type == 2:
                    self.out_channel = self.hidden_channel
                else:
                    self.out_channel = self.hidden_channel
                self.num_linear -= 1
            
            if self.fusion_type == 1:
                raise NotImplementedError

            if self.fusion_type in (0, 2) and self.num_conv == 0:
                self.out_channel *= 2

            if self.num_linear > 0:
                self.linear_layers = torch.nn.Sequential()
                for i in range(self.num_linear):
                    self.linear_layers.add_module(
                        'linear{}'.format(i), torch.nn.Linear(self.out_channel, self.hidden_channel)
                    )
                    if not force_linear:
                        self.linear_layers.add_module(
                            'relu{}'.format(i), torch.nn.ReLU()
                        )
                    self.out_channel = self.hidden_channel
        else:
            self.in_channel = cfg.RCNN.POOL_CHANNEL
            self.out_channel = 6
            self.z_cls_1x1_conv = torch.nn.Conv2d(in_channels=self.in_channel, out_channels=self.in_channel * 2, kernel_size=1)
            self.x_cls_1x1_conv = torch.nn.Conv2d(in_channels=self.in_channel, out_channels=self.in_channel, kernel_size=1)
            self.z_box_1x1_conv = torch.nn.Conv2d(in_channels=self.in_channel, out_channels=self.in_channel * 4, kernel_size=1)
            self.x_box_1x1_conv = torch.nn.Conv2d(in_channels=self.in_channel, out_channels=self.in_channel, kernel_size=1)
            self.bbox_refine = torch.nn.Conv2d(in_channels=4, out_channels=4, kernel_size=1)

    def forward(self, z_feat, x_feat):
        num_x = x_feat.size(0)
        num_z, zc, zh, zw = z_feat.size()
        
        if self.fusion_type != 4:
            if self.fusion_type in (2, 3):
                z_feat = z_feat.view(z_feat.size(0), -1)
                x_feat = x_feat.view(x_feat.size(0), -1)
                z_pre_linear = self.pre_linear(z_feat)
                x_pre_linear = self.pre_linear(x_feat)
                
                if num_z == 1 and num_x > 1:
                    z_pre_linear = z_pre_linear.expand(num_x, -1)

                if self.fusion_type == 2:
                    fusion_feat = torch.cat((z_pre_linear, x_pre_linear), dim=1).contiguous()
                elif self.fusion_type == 3:
                    fusion_feat = (z_pre_linear - x_pre_linear).contiguous()

            elif self.fusion_type == 0:
                if num_z == 1 and num_x > 1:
                    fusion_feat = torch.cat((z_feat.expand(num_x, -1, -1, -1), x_feat), dim=1).contiguous()
                else:
                    fusion_feat = torch.cat((z_feat, x_feat), dim=1).contiguous()
                if self.num_conv > 0:
                    fusion_feat = self.roi_conv(fusion_feat)
                fusion_feat = fusion_feat.view(fusion_feat.size(0), -1)
            elif self.fusion_type == 1:
                fusion_feat = z_feat - x_feat
                fusion_feat = fusion_feat.view(fusion_feat.size(0), -1)
            
            if self.num_linear > 0:
                fusion_feat = self.linear_layers(fusion_feat)
        else:
            z_cls_branch = self.z_cls_1x1_conv(z_feat)
            z_bbox_branch = self.z_box_1x1_conv(z_feat)
            x_cls_branch = self.x_cls_1x1_conv(x_feat)
            x_bbox_branch = self.x_box_1x1_conv(x_feat)
            # cross-correlation

            if num_z == 1 and num_x > 1:
                z_bbox_branch = z_bbox_branch.repeat(num_x, 1, 1, 1).contiguous()
                z_cls_branch = z_cls_branch.repeat(num_x, 1, 1, 1).contiguous()

            rpn_cls_logit = xcorr(z_cls_branch, x_cls_branch, group=2) * 0.001
            rpn_bbox_delta = xcorr(z_bbox_branch, x_bbox_branch, group=4) * 0.001
            rpn_bbox_delta = self.bbox_refine(rpn_bbox_delta)
            
            fusion_feat = torch.cat((rpn_cls_logit, rpn_bbox_delta), dim=1)
            fusion_feat = fusion_feat.view(fusion_feat.size(0), -1)

        return fusion_feat


class _INDENTITY_MAP(torch.nn.Module):

    def __init__(self, s, t):
        super(_INDENTITY_MAP, self).__init__()
        self.s = s
        self.t = t

    def forward(self, x):
        return x[:, self.s:self.t].contiguous()
        

class SiamRCNNHead(torch.nn.Module):

    def __init__(self, cfg, is_training=False):
        super(SiamRCNNHead, self).__init__()

        self.feat_extractor = FeatExtractor(cfg, is_training=is_training)
        self.siam_fusion = SiamFeatFusion(cfg, is_training=is_training)
        
        self.num_classes = cfg.RCNN.NUM_CLASSES
        if self.siam_fusion.fusion_type == 4:
            self.rcnn_cls = _INDENTITY_MAP(0, 2)
            self.rcnn_bbox = _INDENTITY_MAP(2, 6)
        else:
            self.rcnn_cls = torch.nn.Linear(self.siam_fusion.out_channel, self.num_classes)
            self.rcnn_bbox = torch.nn.Linear(self.siam_fusion.out_channel, 4)

        self.is_training = is_training

        if is_training:
            if self.siam_fusion.fusion_type != 4:
                # init weights
                torch.nn.init.xavier_uniform_(self.rcnn_cls.weight)
                torch.nn.init.xavier_uniform_(self.rcnn_bbox.weight)
                self.rcnn_cls.weight.data *= 0.1
                self.rcnn_bbox.weight.data *= 0.01


    def forward(self, z_out, x_out, z_rois, x_rois):
        
        if self.is_training:

            z_feats = self.feat_extractor(z_out, z_rois)
            x_feats = self.feat_extractor(x_out, x_rois)
            fusion_feat = self.siam_fusion(z_feats, x_feats)

            cls_logit = self.rcnn_cls(fusion_feat)
            bbox_delta = self.rcnn_bbox(fusion_feat)

            return cls_logit, bbox_delta

        else:
            raise NotImplementedError
    