import torch
import torch.nn.functional as F
import numpy as np
from copy import deepcopy
from collections import OrderedDict

from models.base_conv import model_factory
from models.siam_rpn_head import SiamRPNHead
from utils.crop import center_crop_tensor

class SiamRPN(torch.nn.Module):

    def __init__(self, cfg, is_training=True):
        
        super(SiamRPN, self).__init__()

        self.cfg = cfg
        self.is_training = is_training


        # construct backbone network
        backbone = model_factory(self.cfg.SIAMRPN.BACKBONE)
        self.z_net = backbone(padding=True, all_layer_out=True)
        if not cfg.SIAMRPN.SHARE_WEIGHTS:
            self.x_net = backbone(padding=True, all_layer_out=True)
        if self.is_training and cfg.SIAMRPN.FREEZE_CONV_NUM > 0:
            self.z_net.freeze_conv_layer(cfg.SIAMRPN.FREEZE_CONV_NUM)
            if not cfg.SIAMRPN.SHARE_WEIGHTS:
                self.x_net.freeze_conv_layer(cfg.SIAMRPN.FREEZE_CONV_NUM)


        self.rpn_head = SiamRPNHead(self.cfg, is_training)
        
        # BGR ImageNet means
        self.means = torch.nn.Parameter(torch.Tensor([104.0,117.0,123.0]).reshape(1, 3, 1, 1), requires_grad = False)

        if cfg.SIAMRPN.RPN_HEAD_LOSS == 'softmax':
            self.use_softmax = True
            self.nc_score_out = len(cfg.SIAMRPN.ANCHOR_SCALES) * len(cfg.SIAMRPN.ANCHOR_RATIOS) * 2
        else:
            self.use_softmax = False # sigmoid
            self.nc_score_out = len(cfg.SIAMRPN.ANCHOR_SCALES) * len(cfg.SIAMRPN.ANCHOR_RATIOS)
        self.nc_bbox_out = len(cfg.SIAMRPN.ANCHOR_SCALES) * len(cfg.SIAMRPN.ANCHOR_RATIOS) * 4

        self.embd_name = cfg.SIAMRPN.EMBEDDING_TENSOR_NAME
        if not is_training:
            self.z_size = self.cfg.INFERENCE.Z_RESPONSE_SIZE
            self.x_size = self.cfg.INFERENCE.X_RESPONSE_SIZE
        else:
            self.z_size = self.cfg.TRAIN.Z_RESPONSE_SIZE
            self.x_size = self.cfg.TRAIN.X_RESPONSE_SIZE


    def forward(self, z_image=None, x_image=None, ar_index=None):

        if z_image is not None:
            num = z_image.size(0)
            z_image = z_image - self.means
            z_out = self.z_net(z_image)
            z_embeddings = center_crop_tensor(z_out[self.embd_name], self.z_size, self.z_size)
            # z_embeddings = self.z_net(z_image)

        else:
            z_out = None
            z_embeddings = None
        
        if x_image is not None:
            x_image = x_image - self.means
            if not self.cfg.SIAMRPN.SHARE_WEIGHTS:
                x_out = self.x_net(x_image)
                # x_embeddings = self.x_net(x_image)
            else:
                x_out = self.z_net(x_image)
                # x_embeddings = self.z_net(x_image)
            x_embeddings = center_crop_tensor(x_out[self.embd_name], self.x_size, self.x_size)
        else:
            x_out = None
            x_embeddings = None
        rpn_cls_logit, rpn_bbox_delta = \
            self.rpn_head(z_embedding=z_embeddings, x_embedding=x_embeddings, ar_index=ar_index)

        # if x_image is None, it means that only forwarding z_image is reset the template features in rpn_head.
        if x_image is None:
            return z_out

        return z_out, x_out, rpn_cls_logit, rpn_bbox_delta
