import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from copy import deepcopy

from utils.misc import xcorr
from utils.model_weights import random_init_weights

class SiamRPNHead(torch.nn.Module):

    def __init__(self, cfg, is_training=True):

        super(SiamRPNHead, self).__init__()

        self.cfg = cfg
        self.is_training = is_training

        # number of output channels
        self.num_channels = self.cfg.SIAMRPN.OUTPUT_CONV_CHANNELS
        self.num_anchors = len(self.cfg.SIAMRPN.ANCHOR_SCALES) * \
                           len(self.cfg.SIAMRPN.ANCHOR_RATIOS)

        if self.cfg.SIAMRPN.RPN_HEAD_LOSS == 'sigmoid':
            self.cls_group = 1
        elif self.cfg.SIAMRPN.RPN_HEAD_LOSS == 'softmax':
            self.cls_group = 2
        else:
            raise ValueError

        self.z_cls_conv = nn.Conv2d(self.num_channels, self.num_channels * self.num_anchors * self.cls_group, 3)
        self.z_bbox_conv = nn.Conv2d(self.num_channels, self.num_channels * self.num_anchors * 4, 3)

        self.x_cls_conv = nn.Sequential(nn.Conv2d(self.num_channels, self.num_channels, 3))
        self.x_bbox_conv = nn.Sequential(nn.Conv2d(self.num_channels, self.num_channels, 3))
        
        self.bbox_refine = nn.Conv2d(self.num_anchors * 4, self.num_anchors * 4, 1)

        # define the aspect ratio embedding function
        self.num_ar_range = len(cfg.SIAMRPN.ASPECT_RATIO_RANGE) + 1

        self.z_size = self.cfg.INFERENCE.Z_RESPONSE_SIZE

        ar_att_dict = {
            'none': 0,
            'cls_channel': 1,
            'cls_spatial': 2
        }
        self.ar_type = ar_att_dict[cfg.SIAMRPN.AR_ATTENTION]
        if self.ar_type == 0:
            pass
        elif self.ar_type == 1:
            self.ar_embeddings = torch.nn.Parameter(
                torch.zeros(self.num_ar_range, self.num_channels * self.num_anchors * self.cls_group), 
                requires_grad=True)
        elif self.ar_type == 2:
            _sz = self.z_size - 2
            self.ar_embeddings = torch.nn.Parameter(
                torch.zeros(self.num_ar_range, _sz * _sz * self.num_anchors * self.cls_group),
                requires_grad=True)
        else:
            raise ValueError('unknown value cfg.SIAMRPN.AR_ATTENTION: {}'.format(self.cfg.SIAMRPN.AR_ATTENTION))

        self.cls_scale_factor = self.cfg.SIAMRPN.CLS_SCALE_FACTOR
        self.bbox_scale_factor = self.cfg.SIAMRPN.BBOX_SCALE_FACTOR
        # Since the bbox branch has an refine step, we don't need to add another bias term in bbox branch
        self.cls_bias = nn.Parameter(torch.zeros(1, self.num_anchors * self.cls_group, 1, 1), requires_grad=True)

        if not is_training:
            self.z_size = self.cfg.INFERENCE.Z_RESPONSE_SIZE
            self.z_cls_weights = nn.Parameter(torch.zeros(self.num_anchors * self.cls_group, self.num_channels, self.z_size-2, self.z_size-2), requires_grad=False)
            self.z_bbox_weights = nn.Parameter(torch.zeros(self.num_anchors * 4, self.num_channels, self.z_size-2, self.z_size-2), requires_grad=False)
        
        
    def forward(self, z_embedding=None, x_embedding=None, ar_index=None):
        if z_embedding is not None:
            z_cls_branch = self.z_cls_conv(z_embedding) # N x 256 * 3 x 4 x 4
            z_bbox_branch = self.z_bbox_conv(z_embedding) # N x 256 * 12 x 4 x 4
            if self.ar_type == 1:
                ar_att = self.ar_embeddings[ar_index, :].contiguous()
                ar_att = ar_att.view(1, self.num_channels * self.num_anchors * self.cls_group, 1, 1)
                z_cls_branch = z_cls_branch * (F.sigmoid(ar_att) + 0.5)
            elif self.ar_type == 2:
                n, c, h, w = z_cls_branch.size()
                z_cls_branch = z_cls_branch.view(n, self.num_anchors * self.cls_group, -1, h, w)
                ar_att = self.ar_embeddings[ar_index, :].contiguous()
                ar_att = ar_att.view(n, self.num_anchors * self.cls_group, 1, self.z_size-2, self.z_size-2)
                z_cls_branch = z_cls_branch * (F.sigmoid(ar_att))
                z_cls_branch = z_cls_branch.view(n, c, h, w)
            else:
                pass

            if not self.is_training:
                assert z_embedding.shape[0] == 1, 'Support batch=1 only.'
                self.z_cls_weights.data[:] = z_cls_branch.view(self.z_cls_weights.size()) * self.cls_scale_factor
                self.z_bbox_weights.data[:] = z_bbox_branch.view(self.z_bbox_weights.size()) * self.bbox_scale_factor
                return None, None

        if x_embedding is not None:
            x_cls_branch = self.x_cls_conv(x_embedding)
            x_bbox_branch = self.x_bbox_conv(x_embedding)

        if self.is_training:
            # correlation
            rpn_cls_logit = xcorr(z_cls_branch, x_cls_branch, group=self.num_anchors * self.cls_group) * self.cls_scale_factor + self.cls_bias
            rpn_bbox_delta = xcorr(z_bbox_branch, x_bbox_branch, group=self.num_anchors * 4) * self.bbox_scale_factor
            rpn_bbox_delta = self.bbox_refine(rpn_bbox_delta)
        else:
            rpn_cls_logit = F.conv2d(x_cls_branch, self.z_cls_weights) + self.cls_bias
            rpn_bbox_delta = F.conv2d(x_bbox_branch, self.z_bbox_weights)
            rpn_bbox_delta = self.bbox_refine(rpn_bbox_delta)

        return rpn_cls_logit, rpn_bbox_delta



    def _initialize_weights(self):
        random_init_weights(self.modules())
        if self.cfg.SIAMRPN.AR_ATTENTION == 'cls_channel':
            if self.cfg.SIAMRPN.AR_EMBEDDING_INIT == 'zero':
                torch.nn.init.zeros_(self.ar_embeddings)
            elif self.cfg.SIAMRPN.AR_EMBEDDING_INIT == 'normal':
                torch.nn.init.normal_(self.ar_embeddings, mean=0, std=1)
            else:
                raise ValueError