
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import torch
import utils.cython_bbox as cython_bbox

bbox_overlaps = cython_bbox.bbox_overlaps
bbox_overlaps_iob = cython_bbox.bbox_overlaps_iob

BBOX_XFORM_CLIP = np.log(255./ 32.)


def crop_bbox_transform(all_boxes, crop_box, image_size):
    bw, bh = crop_box[2] - crop_box[0], crop_box[3] - crop_box[1]
    scale_x, scale_y = float(image_size) / bw, float(image_size) / bh
    shift_x, shift_y = -crop_box[0], -crop_box[1]
    all_boxes[:, 0:4:2] += shift_x
    all_boxes[:, 1:4:2] += shift_y
    all_boxes[:, 0:4:2] *= scale_x
    all_boxes[:, 1:4:2] *= scale_y
    return all_boxes


def boxes_area(boxes):
    """Compute the area of an array of boxes."""
    w = (boxes[:, 2] - boxes[:, 0])
    h = (boxes[:, 3] - boxes[:, 1])
    areas = w * h
    assert np.all(areas >= 0), 'Negative areas founds'
    return areas


def round_and_int(boxes):
    if isinstance(boxes, (list, tuple)):
        assert not isinstance(boxes[0], (list, tuple))
        return [int(round(_)) for _ in boxes]

    elif isinstance(boxes, np.ndarray):
        return np.round(boxes).astype(np.int32)
    elif isinstance(boxes, torch.Tensor):
        return torch.round(boxes).to(dtype=torch.long)
    else:
        raise TypeError('boxes must be a list, tuple, or numpy array.')


def random_shift_boxes(boxes, shift_pixels):

    if isinstance(boxes, (list, tuple)):
        assert len(boxes) == 4
        x_sf = float(np.random.uniform(-shift_pixels, shift_pixels))
        y_sf = float(np.random.uniform(-shift_pixels, shift_pixels))
        return (boxes[0] + x_sf, boxes[1] + y_sf, boxes[2] + x_sf, boxes[3] + x_sf)

    elif isinstance(boxes, np.ndarray):
        num_boxes = boxes.shape[0]
        x_sf = np.random.uniform(-shift_pixels, shift_pixels, size=(num_boxes, 1))
        y_sf = np.random.uniform(-shift_pixels, shift_pixels, size=(num_boxes, 1))
        xyxy = boxes.copy()
        xyxy[:, 0::2] = boxes[:, 0::2] + x_sf
        xyxy[:, 1::2] = boxes[:, 1::2] + y_sf
        return xyxy

    else:
        raise TypeError('boxes must be a list, tuple, or numpy array.')


def xcycwh_to_xyxy(xcycwh):
    """Convert [x_c y_c w h] box format to [x1, y1, x2, y2] format.""" 
    if isinstance(xcycwh, (list, tuple)):
        # Single box given as a list of coordinates
        assert not isinstance(xcycwh[0], (list, tuple))
        xc, yc = xcycwh[0], xcycwh[1]
        w = xcycwh[2]
        h = xcycwh[3]
        x1 = xc - w / 2.
        y1 = yc - h / 2.
        x2 = xc + w / 2.
        y2 = yc + h / 2.
        return [x1, y1, x2, y2]
    elif isinstance(xcycwh, np.ndarray):
        # Multiple boxes given as a 2D ndarray
        if len(xcycwh.shape) == 1:
            wh = xcycwh[2:4]
            x1y1 = xcycwh[0:2] - wh / 2.
            x2y2 = xcycwh[0:2] + wh / 2.
            return np.concatenate((x1y1, x2y2))
        wh = xcycwh[:, 2:4]
        x1y1 = xcycwh[:, 0:2] - wh / 2.
        x2y2 = xcycwh[:, 0:2] + wh / 2.
        return np.hstack((x1y1, x2y2))
    elif isinstance(xcycwh, torch.Tensor):
        wh = xcycwh[:, 2:4]
        x1y1 = xcycwh[:, 0:2] - wh / 2.
        x2y2 = xcycwh[:, 0:2] + wh / 2.
        return torch.cat((x1y1, x2y2), dim=1)
    else:
        raise TypeError('Argument xcycwh must be a list, tuple, or numpy array.')


def xyxy_to_xcycwh(xyxy):
    """Convert [x1 y1 x2, y2] box format to [x_c y_c w h] format."""
    if isinstance(xyxy, (list, tuple)):
        # Single box given as a list of coordinates
        assert not isinstance(xyxy[0], (list, tuple))
        x1, y1 = xyxy[0], xyxy[1]
        w = xyxy[2] - x1
        h = xyxy[3] - y1
        x = (xyxy[0] + xyxy[2]) / 2.
        y = (xyxy[1] + xyxy[3]) / 2.
        return [x, y, w, h]
    elif isinstance(xyxy, np.ndarray):
        if len(xyxy.shape) == 1:
            wh = xyxy[2:4] - xyxy[0:2]
            xy = (xyxy[0:2] + xyxy[2:4]) / 2.
            return np.concatenate((xy, wh))
        # Multiple boxes given as a 2D ndarray
        wh = xyxy[:, 2:4] - xyxy[:, 0:2]
        xy = (xyxy[:, 0:2] + xyxy[:, 2:4]) / 2.
        return np.hstack((xy, wh))
    elif isinstance(xyxy, torch.Tensor):
        wh = xyxy[:, 2:4] - xyxy[:, 0:2]
        xy = (xyxy[:, 0:2] + xyxy[:, 2:4]) / 2.
        return torch.cat((xy, wh), dim=1)
    else:
        raise TypeError('Argument xyxy must be a list, tuple, or numpy array.')

def xywh_to_xyxy(xywh):
    """Convert [x1 y1 w h] box format to [x1 y1 x2 y2] format."""
    if isinstance(xywh, (list, tuple)):
        # Single box given as a list of coordinates
        assert not isinstance(xywh[0], (list, tuple))
        x1, y1 = xywh[0], xywh[1]
        x2 = x1 + np.maximum(0., xywh[2])
        y2 = y1 + np.maximum(0., xywh[3])
        return [x1, y1, x2, y2]
    elif isinstance(xywh, np.ndarray):
        # Multiple boxes given as a 2D ndarray
        return np.hstack(
            (xywh[:, 0:2], xywh[:, 0:2] + np.maximum(0, xywh[:, 2:4]))
        )
    elif isinstance(xywh, torch.Tensor):
        xyxy = xywh.clone()
        xywh[:, 2:4] += xywh[:, 0:2]
    else:
        raise TypeError('Argument xyxy must be a list, tuple, numpy array, or tensor.')


def xyxy_to_xywh(xyxy):
    """Convert [x1 y1 x2 y2] box format to [x1 y1 w h] format."""
    if isinstance(xyxy, (list, tuple)):
        # Single box given as a list of coordinates
        assert not isinstance(xyxy[0], (list, tuple))
        x1, y1 = xyxy[0], xyxy[1]
        w = xyxy[2] - x1 
        h = xyxy[3] - y1
        return [x1, y1, w, h]
    elif isinstance(xyxy, np.ndarray):
        # Multiple boxes given as a 2D ndarray
        return np.hstack((xyxy[:, 0:2], xyxy[:, 2:4] - xyxy[:, 0:2]))
    elif isinstance(xyxy, torch.Tensor):
        xywh = xyxy.clone()
        xywh[:, 2:4] -= xywh[:, 0:2]
        return xywh
    else:
        raise TypeError('Argument xyxy must be a list, tuple, numpy array, or tensor.')

def filter_small_boxes(boxes, min_size):
    """Keep boxes with width and height both greater than min_size."""
    w = boxes[:, 2] - boxes[:, 0]
    h = boxes[:, 3] - boxes[:, 1]
    
    if isinstance(boxes, torch.Tensor):
        valid_inds = torch.nonzero((w > min_size) & (h > min_size)).view(-1)
        kept_boxes = torch.index_select(boxes, dim=0, index=valid_inds)
    else:
        valid_inds = np.where((w > min_size) & (h > min_size))[0]
        kept_boxes = boxes[valid_inds, :]
    return keep_inds


def filter_box_not_in_image(boxes, image_width, image_height):
    wh = boxes[:, 2:4] - boxes[:, 0:2]
    xy = (boxes[:, 0:2] + wh / 2.)
    if isinstance(boxes, torch.Tensor):
        valid_inds = torch.nonzero((xy[:, 0] >= 0) & 
                                (xy[:, 0]  < image_width) &
                                (xy[:, 1] >= 0) & 
                                (xy[:, 1] < image_height)).view(-1)
        new_boxes = torch.index_select(boxes, dim=0, index=valid_inds)
    else:
        valid_inds = np.where((xy[:, 0] >= 0) &
                            (xy[:, 0]  < image_width) &
                            (xy[:, 1] >= 0) &
                            (xy[:, 1] < image_height))[0]
        new_boxes = boxes[valid_inds, :]
    return new_boxes

def bbox_transform_batch(boxes, deltas, batch_size, weights=(1.0, 1.0, 1.0, 1.0)):
    is_torch = isinstance(boxes, torch.Tensor)
    assert is_torch, "bbox_transform_batch only supports torch.Tensor currently."
    widths = boxes[:, :, 2] - boxes[:, :, 0]
    heights = boxes[:, :, 3] - boxes[:, :, 1]
    ctr_x = boxes[:, :, 0] + 0.5 * widths
    ctr_y = boxes[:, :, 1] + 0.5 * heights

    wx, wy, ww, wh = weights
    dx = deltas[:, :, 0::4] / wx
    dy = deltas[:, :, 1::4] / wy
    dw = deltas[:, :, 2::4] / wx
    dh = deltas[:, :, 3::4] / wy

    # Prevent sending too large values into np.exp()

    dw_invalid = torch.isnan(dw)
    dh_invalid = torch.isnan(dh)
    if dw_invalid.any():
        dw[dw_invalid] = 0
    if dh_invalid.any():
        dh[dh_invalid] = 0
    dw = torch.clamp(dw, min=-BBOX_XFORM_CLIP, max=BBOX_XFORM_CLIP)
    dh = torch.clamp(dh, min=-BBOX_XFORM_CLIP, max=BBOX_XFORM_CLIP)


    pred_ctr_x = dx * widths.unsqueeze(2) + ctr_x.unsqueeze(2)
    pred_ctr_y = dy * heights.unsqueeze(2) + ctr_y.unsqueeze(2)
    pred_w = torch.exp(dw) * widths.unsqueeze(2)
    pred_h = torch.exp(dh) * heights.unsqueeze(2)

    pred_boxes = deltas.clone()
    # x1
    pred_boxes[:, :, 0::4] = pred_ctr_x - 0.5 * pred_w
    # y1
    pred_boxes[:, :, 1::4] = pred_ctr_y - 0.5 * pred_h
    # x2
    pred_boxes[:, :, 2::4] = pred_ctr_x + 0.5 * pred_w
    # y2
    pred_boxes[:, :, 3::4] = pred_ctr_y + 0.5 * pred_h

    return pred_boxes



def bbox_transform(boxes, deltas, weights=(1.0, 1.0, 1.0, 1.0)):
    """Forward transform that maps proposal boxes to predicted ground-truth
    boxes using bounding-box regression deltas. See bbox_transform_inv for a
    description of the weights argument.
    """
    is_torch = isinstance(boxes, torch.Tensor)

    if len(boxes) == 0:
        if is_torch:
            return torch.zeros_like(boxes)
        else:
            return np.zeros((0, deltas.shape[1]), dtype=deltas.dtype)
    
    if not is_torch:
        boxes = boxes.astype(deltas.dtype, copy=False)
    else:
        boxes = boxes.to(dtype=deltas.dtype, device=deltas.device)
    
    widths = boxes[:, 2] - boxes[:, 0] + 1.0
    heights = boxes[:, 3] - boxes[:, 1] + 1.0
    ctr_x = boxes[:, 0] + 0.5 * widths
    ctr_y = boxes[:, 1] + 0.5 * heights

    wx, wy, ww, wh = weights
    dx = deltas[:, 0::4] / wx
    dy = deltas[:, 1::4] / wy
    dw = deltas[:, 2::4] / ww
    dh = deltas[:, 3::4] / wh
    
    # Prevent sending too large values into np.exp()
    if not is_torch:
        dw_invalid = (np.isnan(dw) | np.isinf(dw))
        dh_invalid = (np.isnan(dh) | np.isinf(dh))
        dw[dw_invalid] = BBOX_XFORM_CLIP
        dh[dh_invalid] = BBOX_XFORM_CLIP
        dw = np.minimum(dw, BBOX_XFORM_CLIP)
        dh = np.minimum(dh, BBOX_XFORM_CLIP)

        pred_ctr_x = dx * widths[:, np.newaxis] + ctr_x[:, np.newaxis]
        pred_ctr_y = dy * heights[:, np.newaxis] + ctr_y[:, np.newaxis]
        pred_w = np.exp(dw) * widths[:, np.newaxis]
        pred_h = np.exp(dh) * heights[:, np.newaxis]

        pred_boxes = np.zeros(deltas.shape, dtype=deltas.dtype)
        # x1
        pred_boxes[:, 0::4] = pred_ctr_x - 0.5 * pred_w
        # y1
        pred_boxes[:, 1::4] = pred_ctr_y - 0.5 * pred_h
        # x2 (note: "- 1" is correct; don't be fooled by the asymmetry) (removed -1 2018.09.28)
        pred_boxes[:, 2::4] = pred_ctr_x + 0.5 * pred_w 
        # y2 (note: "- 1" is correct; don't be fooled by the asymmetry) (removed -1 2018.09.28)
        pred_boxes[:, 3::4] = pred_ctr_y + 0.5 * pred_h 

    else:
        dw_invalid = torch.isnan(dw)
        dh_invalid = torch.isnan(dh)
        dw[dw_invalid] = 0
        dh[dh_invalid] = 0
        dw = torch.clamp(dw, min=-BBOX_XFORM_CLIP, max=BBOX_XFORM_CLIP)
        dh = torch.clamp(dh, min=-BBOX_XFORM_CLIP, max=BBOX_XFORM_CLIP)
        
        pred_ctr_x = dx * widths.unsqueeze(1) + ctr_x.unsqueeze(1)
        pred_ctr_y = dy * heights.unsqueeze(1) + ctr_y.unsqueeze(1)
        pred_w = torch.exp(dw) * widths.unsqueeze(1)
        pred_h = torch.exp(dh) * heights.unsqueeze(1)

        pred_boxes = torch.zeros_like(deltas)
        # x1
        pred_boxes[:, 0::4] = pred_ctr_x - 0.5 * pred_w
        # y1
        pred_boxes[:, 1::4] = pred_ctr_y - 0.5 * pred_h
        # x2 (note: "- 1" is correct; don't be fooled by the asymmetry) (removed -1 2018.09.28)
        pred_boxes[:, 2::4] = pred_ctr_x + 0.5 * pred_w 
        # y2 (note: "- 1" is correct; don't be fooled by the asymmetry) (removed -1 2018.09.28)
        pred_boxes[:, 3::4] = pred_ctr_y + 0.5 * pred_h 

    return pred_boxes


def bbox_transform_inv(boxes, gt_boxes, weights=(1.0, 1.0, 1.0, 1.0)):
    """Inverse transform that computes target bounding-box regression deltas
    given proposal boxes and ground-truth boxes. The weights argument should be
    a 4-tuple of multiplicative weights that are applied to the regression
    target.
    In older versions of this code (and in py-faster-rcnn), the weights were set
    such that the regression deltas would have unit standard deviation on the
    training dataset. Presently, rather than computing these statistics exactly,
    we use a fixed set of weights (10., 10., 5., 5.) by default. These are
    approximately the weights one would get from COCO using the previous unit
    stdev heuristic.
    """
    is_torch = isinstance(boxes, torch.Tensor)
    
    ex_widths = boxes[:, 2] - boxes[:, 0]
    ex_heights = boxes[:, 3] - boxes[:, 1]
    ex_ctr_x = boxes[:, 0] + 0.5 * ex_widths
    ex_ctr_y = boxes[:, 1] + 0.5 * ex_heights

    if is_torch:
        gt_widths = torch.clamp(gt_boxes[:, 2] - gt_boxes[:, 0], min=1e-3)
        gt_heights = torch.clamp(gt_boxes[:, 3] - gt_boxes[:, 1], min=1e-3)
    else:
        gt_widths = np.maximum(gt_boxes[:, 2] - gt_boxes[:, 0], 1e-3)
        gt_heights = np.maximum(gt_boxes[:, 3] - gt_boxes[:, 1], 1e-3)
    
    gt_ctr_x = gt_boxes[:, 0] + 0.5 * gt_widths
    gt_ctr_y = gt_boxes[:, 1] + 0.5 * gt_heights

    wx, wy, ww, wh = weights
    targets_dx = wx * (gt_ctr_x - ex_ctr_x) / ex_widths
    targets_dy = wy * (gt_ctr_y - ex_ctr_y) / ex_heights

    if is_torch:
        targets_dw = ww * torch.log(gt_widths / ex_widths)
        targets_dh = wh * torch.log(gt_heights / ex_heights)
        targets = torch.cat((targets_dx.unsqueeze(1), 
                             targets_dy.unsqueeze(1), 
                             targets_dw.unsqueeze(1),
                             targets_dh.unsqueeze(1)), dim=1)

    else:
        targets_dw = ww * np.log(gt_widths / ex_widths)
        targets_dh = wh * np.log(gt_heights / ex_heights)

        targets = np.vstack((targets_dx, targets_dy, targets_dw,
                            targets_dh)).transpose()
    return targets


def bbox_transform_inv_batch(ex_rois, gt_rois, weights=(1.0, 1.0, 1.0, 1.0)):

    if ex_rois.dim() == 2:
        ex_widths = ex_rois[:, 2] - ex_rois[:, 0] + 1.0
        ex_heights = ex_rois[:, 3] - ex_rois[:, 1] + 1.0
        ex_ctr_x = ex_rois[:, 0] + 0.5 * ex_widths
        ex_ctr_y = ex_rois[:, 1] + 0.5 * ex_heights

        gt_widths = gt_rois[:, :, 2] - gt_rois[:, :, 0] + 1.0
        gt_heights = gt_rois[:, :, 3] - gt_rois[:, :, 1] + 1.0
        gt_ctr_x = gt_rois[:, :, 0] + 0.5 * gt_widths
        gt_ctr_y = gt_rois[:, :, 1] + 0.5 * gt_heights
        wx, wy, ww, wh = weights
        targets_dx = (gt_ctr_x - ex_ctr_x.view(1,-1).expand_as(gt_ctr_x)) / ex_widths * wx
        targets_dy = (gt_ctr_y - ex_ctr_y.view(1,-1).expand_as(gt_ctr_y)) / ex_heights * wy
        targets_dw = torch.log(gt_widths / ex_widths.view(1,-1).expand_as(gt_widths)) * ww
        targets_dh = torch.log(gt_heights / ex_heights.view(1,-1).expand_as(gt_heights)) * wh

    elif ex_rois.dim() == 3:
        ex_widths = ex_rois[:, :, 2] - ex_rois[:, :, 0]
        ex_heights = ex_rois[:,:, 3] - ex_rois[:,:, 1]
        ex_ctr_x = ex_rois[:, :, 0] + 0.5 * ex_widths
        ex_ctr_y = ex_rois[:, :, 1] + 0.5 * ex_heights

        gt_widths = gt_rois[:, :, 2] - gt_rois[:, :, 0]
        gt_heights = gt_rois[:, :, 3] - gt_rois[:, :, 1]
        gt_ctr_x = gt_rois[:, :, 0] + 0.5 * gt_widths
        gt_ctr_y = gt_rois[:, :, 1] + 0.5 * gt_heights

        wx, wy, ww, wh = weights
        targets_dx = (gt_ctr_x - ex_ctr_x) / ex_widths * wx
        targets_dy = (gt_ctr_y - ex_ctr_y) / ex_heights * wy
        targets_dw = torch.log(gt_widths / ex_widths) * ww
        targets_dh = torch.log(gt_heights / ex_heights) * wh
    else:
        raise ValueError('ex_roi input dimension is not correct.')

    targets = torch.stack(
        (targets_dx, targets_dy, targets_dw, targets_dh),2)

    return targets




def bbox_overlaps_batch(anchors, gt_boxes):
    """
    anchors: (N, 4) ndarray of float
    gt_boxes: (b, K, 5) ndarray of float
    overlaps: (N, K) ndarray of overlap between boxes and query_boxes
    """
    batch_size = gt_boxes.size(0)


    if anchors.dim() == 2:

        N = anchors.size(0)
        K = gt_boxes.size(1)

        anchors = anchors.view(1, N, 4).expand(batch_size, N, 4).contiguous()
        gt_boxes = gt_boxes[:,:,:4].contiguous()


        gt_boxes_x = (gt_boxes[:,:,2] - gt_boxes[:,:,0] + 1)
        gt_boxes_y = (gt_boxes[:,:,3] - gt_boxes[:,:,1] + 1)
        gt_boxes_area = (gt_boxes_x * gt_boxes_y).view(batch_size, 1, K)

        anchors_boxes_x = (anchors[:,:,2] - anchors[:,:,0] + 1)
        anchors_boxes_y = (anchors[:,:,3] - anchors[:,:,1] + 1)
        anchors_area = (anchors_boxes_x * anchors_boxes_y).view(batch_size, N, 1)

        gt_area_zero = (gt_boxes_x == 1) & (gt_boxes_y == 1)
        anchors_area_zero = (anchors_boxes_x == 1) & (anchors_boxes_y == 1)

        boxes = anchors.view(batch_size, N, 1, 4).expand(batch_size, N, K, 4)
        query_boxes = gt_boxes.view(batch_size, 1, K, 4).expand(batch_size, N, K, 4)

        iw = (torch.min(boxes[:,:,:,2], query_boxes[:,:,:,2]) -
            torch.max(boxes[:,:,:,0], query_boxes[:,:,:,0]) + 1)
        iw[iw < 0] = 0

        ih = (torch.min(boxes[:,:,:,3], query_boxes[:,:,:,3]) -
            torch.max(boxes[:,:,:,1], query_boxes[:,:,:,1]) + 1)
        ih[ih < 0] = 0
        ua = anchors_area + gt_boxes_area - (iw * ih)
        overlaps = iw * ih / ua

        # mask the overlap here.
        overlaps.masked_fill_(gt_area_zero.view(batch_size, 1, K).expand(batch_size, N, K), 0)
        overlaps.masked_fill_(anchors_area_zero.view(batch_size, N, 1).expand(batch_size, N, K), -1)

    elif anchors.dim() == 3:
        N = anchors.size(1)
        K = gt_boxes.size(1)

        if anchors.size(2) == 4:
            anchors = anchors[:,:,:4].contiguous()
        else:
            anchors = anchors[:,:,1:5].contiguous()

        gt_boxes = gt_boxes[:,:,:4].contiguous()

        gt_boxes_x = (gt_boxes[:,:,2] - gt_boxes[:,:,0])
        gt_boxes_y = (gt_boxes[:,:,3] - gt_boxes[:,:,1])
        gt_boxes_area = (gt_boxes_x * gt_boxes_y).view(batch_size, 1, K)

        anchors_boxes_x = (anchors[:,:,2] - anchors[:,:,0])
        anchors_boxes_y = (anchors[:,:,3] - anchors[:,:,1])
        anchors_area = (anchors_boxes_x * anchors_boxes_y).view(batch_size, N, 1)

        gt_area_zero = (gt_boxes_x == 0) & (gt_boxes_y == 0)
        anchors_area_zero = (anchors_boxes_x == 0) & (anchors_boxes_y == 0)

        boxes = anchors.view(batch_size, N, 1, 4).expand(batch_size, N, K, 4)
        query_boxes = gt_boxes.view(batch_size, 1, K, 4).expand(batch_size, N, K, 4)

        iw = (torch.min(boxes[:,:,:,2], query_boxes[:,:,:,2]) -
            torch.max(boxes[:,:,:,0], query_boxes[:,:,:,0]))
        iw[iw < 0] = 0

        ih = (torch.min(boxes[:,:,:,3], query_boxes[:,:,:,3]) -
            torch.max(boxes[:,:,:,1], query_boxes[:,:,:,1]))
        ih[ih < 0] = 0
        ua = anchors_area + gt_boxes_area - (iw * ih)

        overlaps = iw * ih / ua

        # mask the overlap here.
        overlaps.masked_fill_(gt_area_zero.view(batch_size, 1, K).expand(batch_size, N, K), 0)
        overlaps.masked_fill_(anchors_area_zero.view(batch_size, N, 1).expand(batch_size, N, K), -1)
    else:
        raise ValueError('anchors input dimension is not correct.')

    return overlaps