"""Crop operations"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


import torch
import numpy as np

import utils.boxes as ubox
from utils.crop_and_resize import CropAndResizeFunction

def center_crop_tensor(t, crop_height, crop_width):
    n, c, h, w = t.size()
    sy = (h - crop_height) // 2
    ty = sy + crop_height
    sx = (w - crop_width) // 2
    tx = sx + crop_width
    crop_tensor = t[:, :, sy:ty, sx:tx].contiguous()
    return crop_tensor

def generate_crop_boxes(boxes, context_amount=0.5, keep_ar=True):
    boxes_wh = ubox.xyxy_to_xcycwh(boxes)

    if keep_ar:
        ctx_size = (boxes_wh[:, 2] + boxes_wh[:, 3]) * context_amount
        wc_z = boxes_wh[:, 2] + ctx_size
        hc_z = boxes_wh[:, 3] + ctx_size
        if isinstance(boxes_wh, np.ndarray):
            s_z = np.sqrt(wc_z * hc_z)
        else:
            s_z = torch.sqrt(wc_z * hc_z)
        boxes_wh[:, 2] = s_z
        boxes_wh[:, 3] = s_z
    else:    
        boxes_wh[:, 2] = boxes_wh[:, 2] + boxes_wh[:, 2] * context_amount * 2
        boxes_wh[:, 3] = boxes_wh[:, 3] + boxes_wh[:, 3] * context_amount * 2
    crop_boxes = ubox.xcycwh_to_xyxy(boxes_wh)
    
    return crop_boxes


def generate_search_region_boxes(boxes, size_z=127, size_x=255, context_amount=0.5):
    ''' Generate the crop boxes for tracking 
    
    Parameters:
        boxes: list/numpy.ndarry/torch.Tensor , in order of [x1, y1, x2, y2]
    
    Returns:
        box_z: template crop box, same type with boxes
        box_x: search region crop box, same type with boxes
        scale_x: scale factor to cropped domain
    '''

    boxes_wh = ubox.xyxy_to_xcycwh(boxes)
    if isinstance(boxes, list):
        assert not isinstance(boxes[0], list), 'One dimentional list supported only'
        cx, cy, w, h = boxes_wh
        wc_z = w + context_amount * (w + h)
        hc_z = h + context_amount * (w + h)
        s_z = np.sqrt(wc_z * hc_z)
        scale_z = size_z / s_z

        d_search = (size_x - size_z) / 2
        pad = d_search / scale_z
        s_x = s_z + 2 * pad
        scale_x = size_x / s_x

        box_z_wh = [cx, cy, s_z, s_z]
        box_x_wh = [cx, cy, s_x, s_x]

    elif isinstance(boxes, np.ndarray):
        ctx_size = (boxes_wh[:, 2] + boxes_wh[:, 3]) * context_amount
        wc_z = boxes_wh[:, 2] + ctx_size
        hc_z = boxes_wh[:, 3] + ctx_size
        s_z = np.sqrt(wc_z * hc_z)
        scale_z = size_z / s_z

        d_search = (size_x - size_z) / 2
        pad = d_search / scale_z
        s_x = s_z + 2 * pad
        scale_x = size_x / s_x

        box_z_wh = np.hstack((boxes_wh[:, :2], s_z[:, np.newaxis], s_z[:, np.newaxis]))
        box_x_wh = np.hstack((boxes_wh[:, :2], s_x[:, np.newaxis], s_x[:, np.newaxis]))

    elif isinstance(boxes, torch.Tensor):
        ctx_size = (boxes_wh[:, 2] + boxes_wh[:, 3]) * context_amount
        wc_z = boxes_wh[:, 2] + ctx_size
        hc_z = boxes_wh[:, 3] + ctx_size
        s_z = torch.sqrt(wc_z * hc_z)
        scale_z = size_z / s_z

        d_search = (size_x - size_z) / 2
        pad = d_search / scale_z
        s_x = s_z + 2 * pad
        scale_x = size_x / s_x

        box_z_wh = torch.cat((boxes_wh[:, :2], s_z.view(-1, 1), s_z.view(-1, 1)), dim=1).contiguous()
        box_x_wh = torch.cat((boxes_wh[:, :2], s_x.view(-1, 1), s_x.view(-1, 1)), dim=1).contiguous()

    box_z = ubox.xcycwh_to_xyxy(box_z_wh)
    box_x = ubox.xcycwh_to_xyxy(box_x_wh)

    return box_z, box_x, scale_x


def crop_with_boxes(img_tensor, x_crop_boxes, out_height, out_width, crop_inds = None, avg_channels=True, has_normed_coords=False):
    '''Crop the image tensor by given boxes. The output will be resized to target size
    
    Params:
        img_tensor: torch.Tensor, in shape of [N, C, H, W]. If N > 1, the crop_inds must be specified.
        crop_boxes: list/numpy.ndarray/torch.Tensor in shape of [K x 4].
        out_height: int.
        out_width: int.
        crop_inds: list/numpy.ndarray/torch.Tensor in shape of [K]
    Returns:
        crop_img_tensor: torch.Tensor, in shape of [K, C, H, W]    
    '''

    img_device = img_tensor.device

    if isinstance(x_crop_boxes, list):
        crop_boxes = torch.tensor(x_crop_boxes, dtype=torch.float32).to(img_device)
    elif isinstance(x_crop_boxes, np.ndarray):
        crop_boxes = torch.tensor(x_crop_boxes, dtype=torch.float32).to(img_device)
    elif isinstance(x_crop_boxes, torch.Tensor):
        # change type and device if necessary
        crop_boxes = x_crop_boxes.clone().to(device=img_device, dtype=torch.float32)
    else:
        raise ValueError('Unknown type for crop_boxes {}'.format(type(x_crop_boxes)))

    if len(crop_boxes.size()) == 1:
        crop_boxes = crop_boxes.view(1, 4)

    num_imgs, chanenls, img_height, img_width = img_tensor.size()
    num_crops = crop_boxes.size(0)

    if crop_inds is not None:
        if isinstance(crop_inds, list) or isinstance(crop_inds, np.ndarray):
            crop_inds = torch.tensor(crop_inds, dtype=torch.float32).to(img_device)
        elif isinstance(crop_inds, torch.Tensor):
            crop_inds = crop_inds.to(device=img_device, dtype=torch.float32)
        else:
            raise ValueError('Unknown type for crop_inds {}'.format(type(crop_inds)))
        crop_inds = crop_inds.view(-1)
        assert crop_inds.size(0) == crop_boxes.size(0)
    else:
        if num_imgs == 1:
            crop_inds = torch.zeros(num_crops, dtype=torch.float32, device=img_device)
        elif num_imgs == num_crops:
            crop_inds = torch.arange(num_crops, dtype=torch.float32, device=img_device)
        else:
            raise ValueError('crop_inds MUST NOT be None.')
    
    if avg_channels:
        img_channel_avg = img_tensor.mean(dim=2, keepdim=True).mean(dim=3, keepdim=True)
        img_tensor_minus_avg = img_tensor - img_channel_avg # minus mean values
    
    crop_img_tensor = CropAndResizeFunction(out_height, out_width, has_normed=has_normed_coords)(
                        img_tensor_minus_avg, crop_boxes, crop_inds)
    
    if avg_channels:
        # add mean value
        crop_img_tensor += img_channel_avg[crop_inds.long()]
    
    return crop_img_tensor