import torch
from torch.autograd import Function
from . import crop_and_resize as _B

def crop_and_resize(image, boxes, box_ind, crop_size):
    ''' Crop patches from image by given boxes and resize them to given size.

    Args:
        img: input image. A Tensor of shape [batch, channel, image_height, image_width].
        boxes: A Tensor of type Torch.FloatTensor. A 2D tensor of shape [num_boxes, 4], the i-th row of the tensor
               specifies the coordinates of a box in the box_inds[i] image and is specified in normalized 
               coordinates [y1, x1, y2, x2].
        box_inds:  A Tensor of type Torch.IntTensor. A 1-D tensor of shape [num_boxes] with int32 values in [0, batch).
        crop_size:  A Tuple indicates the output size. size = [crop_height, crop_width]. 
                    All cropped image patches are resized to this size.
    
    Returns:
        A Torch.Tensor of cropped images.

    '''

    # crops = torch.zeros_like(image)
    if image.is_cuda:
        
        with torch.cuda.device(image.device.index):
            crops = _B.forward(image, boxes, box_ind, 0.0, crop_size[0], crop_size[1])

    else:
        crops = _B.forward(image, boxes, box_ind, 0.0, crop_size[0], crop_size[1])

    return crops


class CropAndResizeFunction(Function):

    def __init__(self, crop_height, crop_width, extrapolation_value=0.0, has_normed=False):

        self.crop_height = crop_height
        self.crop_width = crop_width
        self.extrapolation_value = extrapolation_value
        self.has_normed = has_normed

    def forward(self, image, boxes, box_inds):

        box_inds = box_inds.contiguous()
        crop_boxes = boxes[:, :4].contiguous()
        
        n, c, h, w = image.size()
        self.batch_size = int(n)
        self.image_height = int(h)
        self.image_width = int(w)

        if not self.has_normed:
            # normalization
            crop_boxes[:, 0:4:2] /= (w - 1)
            crop_boxes[:, 1:4:2] /= (h - 1)
        
        if image.is_cuda:
            with torch.cuda.device(image.device.index):
                crops = _B.forward(image, crop_boxes, box_inds, 
                    self.extrapolation_value, self.crop_height, self.crop_width)
        else:
            crops = _B.forward(image, crop_boxes, box_inds, 
                    self.extrapolation_value, self.crop_height, self.crop_width)
        
        # save for backward
        self.save_for_backward(crop_boxes, box_inds)

        return crops

    def backward(self, grad_outputs):
        boxes, box_ind = self.saved_tensors
        grad_outputs = grad_outputs.contiguous()
        if grad_outputs.is_cuda:
            with torch.cuda.device(grad_outputs.device.index):
                grads_image = _B.backward(grad_outputs, boxes, box_ind, 
                            self.batch_size, self.image_height, self.image_width)
        else:
            grads_image = _B.backward(grad_outputs, boxes, box_ind, 
                            self.batch_size, self.image_height, self.image_width)
        return grads_image, None, None