import torch
import torch.nn.functional as F
import numpy as np
import os
import datetime
from collections import OrderedDict

import utils.boxes as ubox

def get_center(x):
    return (x - 1.) / 2.


def xcorr_g(z, x, groups=1):
    out = []
    n, c, h, w = z.shape
    for i in range(n):
        out.append(torch.nn.functional.conv2d(x[i,:,:,:].unsqueeze(0), 
                                              z[i,:,:,:].unsqueeze(0).view(groups, c // groups, h, w), 
                                              groups=groups))        
    return torch.cat(out, dim=0)


def xcorr(z, x, group=1):
    out = []
    
    n, c, h, w = z.shape
    
    if group == 1:
        _z = z
    else:
        _z = z.view(n * group, c // group, h, w)

    for i in range(n):
        if group == 1:
            out.append(torch.nn.functional.conv2d(x[i,:,:,:].unsqueeze(0), _z[i,:,:,:].unsqueeze(0)))        
        else:
            out.append(torch.nn.functional.conv2d(x[i,:,:,:].unsqueeze(0), _z[i*group:(i+1)*group,:,:,:]))
    return torch.cat(out, dim=0)


def mkdir_if_not_exists(dir_name):
    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)


def init_seed(worker_id):
    np.random.seed(worker_id + datetime.datetime.now().second * datetime.datetime.now().microsecond)


def rpn_transform_to_boxes(anchors, rpn_cls, rpn_bbox_pred, rpn_head_type='softmax'):
    num_batch = int(rpn_cls.size(0))
    num_anchors = len(anchors)
    anchors_batch = np.tile(anchors, [num_batch, 1])
    with torch.no_grad():
        if rpn_head_type == 'sigmoid':
            rpn_cls_score = F.sigmoid(rpn_cls)
        elif rpn_head_type == 'softmax':
            n, c, h, w = rpn_cls.size()
            rpn_cls_reshape = rpn_reshape(rpn_cls, 2)
            rpn_cls_softmax = F.softmax(rpn_cls_reshape, dim=1).view(n, c, h, w)
            rpn_cls_score = rpn_cls_softmax[:, c // 2:, :, :]
        else:
            raise ValueError
    bbox_delta = rpn_bbox_pred.data.cpu().numpy().transpose((0, 2, 3, 1)).reshape(-1, 4)
    score = rpn_cls_score.data.cpu().numpy().transpose((0, 2, 3, 1)).flatten()
    bbox_pred = ubox.bbox_transform(anchors_batch, bbox_delta)

    det_boxes = np.hstack((bbox_pred, score[:, np.newaxis]))

    det_boxes = det_boxes.reshape(num_batch, num_anchors, 5)
    return det_boxes
    