import torch
import os
from collections import OrderedDict

def random_init_weights(model_modules):
    ''' Random initialize model weights. 
    
    Args:
        model_modules: an iterator of torch.nn.Module.modules()

    Returns:
        None
    '''
    for m in model_modules:
        if isinstance(m, torch.nn.Conv2d):
            output_channel = m.weight.shape[0]
            num_groups = m.groups
            channels_per_group = int(output_channel / num_groups)
            for i in range(num_groups):
                st = channels_per_group * i
                end = channels_per_group * (i+1)
                torch.nn.init.xavier_uniform_(m.weight[st:end, :, :, :])
            if m.bias is not None:
                m.bias.data.zero_()
        if isinstance(m, torch.nn.modules.batchnorm._BatchNorm):
            if m.weight is not None:
                torch.nn.init.constant_(m.weight, 1.0)
            if m.bias is not None:
                torch.nn.init.constant_(m.bias, 0.0)
        if isinstance(m, torch.nn.Linear):
            if m.weight is not None:
                torch.nn.init.xavier_uniform_(m.weight)
            if m.bias is not None:
                m.bias.data.zero_()


def init_weights(model, model_path):
    assert os.path.exists(model_path)
        
    pretrained_weights = torch.load(model_path)
    if hasattr(model, 'z_net'):
        st = model.z_net.state_dict()
        pretrained_st = OrderedDict()

        for k, v in st.items():
            if k in pretrained_weights:
                pretrained_st[k] = pretrained_weights[k]
                print("Init z-net {} from pre-trained model {}".format(k, model_path))
            else:
                pretrained_st[k] = st[k]
                # print("Ignore z-net {}".format(k))
        
        model.z_net.load_state_dict(pretrained_st)

    if hasattr(model, 'x_net'):
        st = model.x_net.state_dict()
        pretrained_st = OrderedDict()

        for k, v in st.items():
            if k in pretrained_weights:
                pretrained_st[k] = pretrained_weights[k]
                print("Init x-net {} from pre-trained model {}".format(k, model_path))
            else:
                pretrained_st[k] = st[k]
                # print("Ignore x-net {}".format(k))
        
        model.x_net.load_state_dict(pretrained_st)

    st = model.state_dict()
    pretrained_st = OrderedDict()
    for k, v in st.items():
        if k in pretrained_weights:
            pretrained_st[k] = pretrained_weights[k]
            print("Init model {} from pre-trained model {}".format(k, model_path))
        else:
            pretrained_st[k] = st[k]
    model.load_state_dict(pretrained_st)


def load_weights(model, model_path, verbose=False):
    st = model.state_dict()
    load_st = OrderedDict()
    weights = torch.load(model_path)
    for k, v in st.items():
        if k in weights:
            if verbose:
                print('Loading {}'.format(k))
            load_st[k] = weights[k]
        else:
            if verbose:
                print('Randomly init {}'.format(k))
            load_st[k] = st[k]
    model.load_state_dict(load_st)
